@@ -1,6 +1,6 @@
 /*
  * Sylpheed -- a GTK+ based, lightweight, and fast e-mail client
- * Copyright (C) 1999-2002 Hiroyuki Yamamoto
+ * Copyright (C) 1999-2005 Hiroyuki Yamamoto
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -26,6 +26,13 @@
 #include <glib.h>
 #include <glib/gi18n.h>
 #include <gdk/gdkkeysyms.h>
+#include <gtk/gtkwindow.h>
+#include <gtk/gtkclist.h>
+#include <gtk/gtkvbox.h>
+#include <gtk/gtkstatusbar.h>
+#include <gtk/gtktable.h>
+#include <gtk/gtkentry.h>
+#include <gtk/gtkstock.h>
 
 #include "addressbook.h"
 #include "addressitem.h"
@@ -59,6 +66,7 @@ typedef enum {
 
 static struct _GroupEdit_dlg {
 	GtkWidget *window;
+	GtkWidget *hbbox;
 	GtkWidget *ok_btn;
 	GtkWidget *cancel_btn;
 	GtkWidget *statusbar;
@@ -223,7 +231,7 @@ static gint edit_group_list_compare_func( GtkCList *clist, gconstpointer ptr1, g
 	if( cell2 ) name2 = cell2->u.text;
 	if( ! name1 ) return ( name2 != NULL );
 	if( ! name2 ) return -1;
-	return strcasecmp( name1, name2 );
+	return g_ascii_strcasecmp( name1, name2 );
 }
 
 static void addressbook_edit_group_create( gboolean *cancelled ) {
@@ -394,6 +402,7 @@ static void addressbook_edit_group_create( gboolean *cancelled ) {
 			 G_CALLBACK(edit_group_list_group_button), NULL);
 
 	groupeditdlg.window     = window;
+	groupeditdlg.hbbox      = hbbox;
 	groupeditdlg.ok_btn     = ok_btn;
 	groupeditdlg.cancel_btn = cancel_btn;
 	groupeditdlg.statusbar  = statusbar;
@@ -438,6 +447,8 @@ ItemGroup *addressbook_edit_group( AddressBookFile *abf, ItemFolder *parent, Ite
 
 	if (!groupeditdlg.window)
 		addressbook_edit_group_create(&cancelled);
+	gtkut_box_set_reverse_order(GTK_BOX(groupeditdlg.hbbox),
+				    !prefs_common.comply_gnome_hig);
 	gtk_widget_grab_focus(groupeditdlg.ok_btn);
 	gtk_widget_grab_focus(groupeditdlg.entry_name);
 	gtk_widget_show(groupeditdlg.window);
@@ -521,7 +532,7 @@ ItemFolder *addressbook_edit_folder( AddressBookFile *abf, ItemFolder *parent, I
 		return NULL;
 	}
 	if( folder ) {
-		if( strcasecmp( name, ADDRITEM_NAME(folder) ) == 0 ) {
+		if( g_ascii_strcasecmp( name, ADDRITEM_NAME(folder) ) == 0 ) {
 			g_free( name );
 			return NULL;
 		}
