@@ -397,7 +397,8 @@ static gboolean filter_match_header_cond(FilterCond *cond, GSList *hlist)
 
 		switch (cond->type) {
 		case FLT_COND_HEADER:
-			if (!strcasecmp(header->name, cond->header_name)) {
+			if (!g_ascii_strcasecmp
+				(header->name, cond->header_name)) {
 				if (!cond->str_value ||
 				    cond->match_func(header->body,
 						     cond->str_value))
@@ -410,8 +411,8 @@ static gboolean filter_match_header_cond(FilterCond *cond, GSList *hlist)
 				matched = TRUE;
 			break;
 		case FLT_COND_TO_OR_CC:
-			if (!strcasecmp(header->name, "To") ||
-			    !strcasecmp(header->name, "Cc")) {
+			if (!g_ascii_strcasecmp(header->name, "To") ||
+			    !g_ascii_strcasecmp(header->name, "Cc")) {
 				if (!cond->str_value ||
 				    cond->match_func(header->body,
 						     cond->str_value))
@@ -1091,6 +1092,8 @@ void filter_rule_rename_dest_path(FilterRule *rule, const gchar *old_path,
 	gchar *dest_path;
 	gint oldpathlen;
 
+	oldpathlen = strlen(old_path);
+
 	for (cur = rule->action_list; cur != NULL; cur = cur->next) {
 		action = (FilterAction *)cur->data;
 
@@ -1098,10 +1101,11 @@ void filter_rule_rename_dest_path(FilterRule *rule, const gchar *old_path,
 		    action->type != FLT_ACTION_COPY)
 			continue;
 
-		oldpathlen = strlen(old_path);
 		if (action->str_value &&
 		    !strncmp(old_path, action->str_value, oldpathlen)) {
 			base = action->str_value + oldpathlen;
+			if (*base != G_DIR_SEPARATOR && *base != '\0')
+				continue;
 			while (*base == G_DIR_SEPARATOR) base++;
 			if (*base == '\0')
 				dest_path = g_strdup(new_path);
@@ -1109,6 +1113,9 @@ void filter_rule_rename_dest_path(FilterRule *rule, const gchar *old_path,
 				dest_path = g_strconcat(new_path,
 							G_DIR_SEPARATOR_S,
 							base, NULL);
+			debug_print("filter_rule_rename_dest_path(): "
+				    "renaming %s -> %s\n",
+				    action->str_value, dest_path);
 			g_free(action->str_value);
 			action->str_value = dest_path;
 		}
@@ -1120,6 +1127,9 @@ void filter_rule_delete_action_by_dest_path(FilterRule *rule, const gchar *path)
 	FilterAction *action;
 	GSList *cur;
 	GSList *next;
+	gint pathlen;
+
+	pathlen = strlen(path);
 
 	for (cur = rule->action_list; cur != NULL; cur = next) {
 		action = (FilterAction *)cur->data;
@@ -1130,7 +1140,11 @@ void filter_rule_delete_action_by_dest_path(FilterRule *rule, const gchar *path)
 			continue;
 
 		if (action->str_value &&
-		    !strncmp(path, action->str_value, strlen(path))) {
+		    !strncmp(path, action->str_value, pathlen) &&
+		    (action->str_value[pathlen] == G_DIR_SEPARATOR ||
+		     action->str_value[pathlen] == '\0')) {
+			debug_print("filter_rule_delete_action_by_dest_path(): "
+				    "deleting %s\n", action->str_value);
 			rule->action_list = g_slist_remove
 				(rule->action_list, action);
 			filter_action_free(action);
@@ -1168,6 +1182,15 @@ void filter_rule_match_type_str_to_enum(const gchar *match_type,
 	}
 }
 
+void filter_rule_list_free(GSList *fltlist)
+{
+	GSList *cur;
+
+	for (cur = fltlist; cur != NULL; cur = cur->next)
+		filter_rule_free((FilterRule *)cur->data);
+	g_slist_free(fltlist);
+}
+
 void filter_rule_free(FilterRule *rule)
 {
 	if (!rule) return;
