@@ -200,6 +200,9 @@ struct _FolderClass
 	gint     (*rename_folder)	(Folder		*folder,
 					 FolderItem	*item,
 					 const gchar	*name);
+	gint     (*move_folder)		(Folder		*folder,
+					 FolderItem	*item,
+					 FolderItem	*new_parent);
 	gint     (*remove_folder)	(Folder		*folder,
 					 FolderItem	*item);
 };
@@ -255,6 +258,9 @@ struct _FolderItem
 	guint opened    : 1; /* opened by summary view */
 	guint updated   : 1; /* folderview should be updated */
 
+	guint cache_dirty : 1; /* cache file needs to be updated */
+	guint mark_dirty  : 1; /* mark file needs to be updated */
+
 	FolderSortKey sort_key;
 	FolderSortType sort_type;
 
