@@ -26,6 +26,7 @@
 #include <gtk/gtkscrolledwindow.h>
 #include <gtk/gtktreestore.h>
 #include <gtk/gtktreeview.h>
+#include <gtk/gtktreeselection.h>
 #include <gtk/gtkcellrendererpixbuf.h>
 #include <gtk/gtkcellrenderertext.h>
 #include <gtk/gtksignal.h>
@@ -34,6 +35,8 @@
 #include <gtk/gtkmenu.h>
 #include <gtk/gtkmenuitem.h>
 #include <gtk/gtkitemfactory.h>
+#include <gtk/gtkstock.h>
+#include <gtk/gtkversion.h>
 #include <stdio.h>
 #include <string.h>
 #include <stdlib.h>
@@ -45,6 +48,7 @@
 #include "summary_search.h"
 #include "inputdialog.h"
 #include "grouplistdialog.h"
+#include "foldersel.h"
 #include "manage_window.h"
 #include "alertpanel.h"
 #include "menu.h"
@@ -82,7 +86,7 @@ enum
 { \
 	gtk_statusbar_push(GTK_STATUSBAR(mainwin->statusbar), \
 			   mainwin->folderview_cid, str); \
-	gtkut_widget_wait_for_draw(mainwin->statusbar); \
+	gtkut_widget_draw_now(mainwin->statusbar); \
 }
 
 #define STATUSBAR_POP(mainwin) \
@@ -98,6 +102,7 @@ static GdkPixbuf *outbox_pixbuf;
 static GdkPixbuf *folder_pixbuf;
 static GdkPixbuf *folderopen_pixbuf;
 static GdkPixbuf *foldernoselect_pixbuf;
+static GdkPixbuf *draft_pixbuf;
 static GdkPixbuf *trash_pixbuf;
 
 static void folderview_select_row	(FolderView	*folderview,
@@ -162,6 +167,9 @@ static void folderview_new_folder_cb	(FolderView	*folderview,
 static void folderview_rename_folder_cb	(FolderView	*folderview,
 					 guint		 action,
 					 GtkWidget	*widget);
+static void folderview_move_folder_cb	(FolderView	*folderview,
+					 guint		 action,
+					 GtkWidget	*widget);
 static void folderview_delete_folder_cb	(FolderView	*folderview,
 					 guint		 action,
 					 GtkWidget	*widget);
@@ -213,10 +221,16 @@ static void folderview_drag_received_cb  (GtkWidget        *widget,
 					  guint             time,
 					  FolderView       *folderview);
 
+static GtkTargetEntry folderview_drag_types[] =
+{
+	{"text/plain", GTK_TARGET_SAME_APP, 0}
+};
+
 static GtkItemFactoryEntry folderview_mail_popup_entries[] =
 {
-	{N_("/Create _new folder..."),	NULL, folderview_new_folder_cb,    0, NULL},
+	{N_("/Create _new folder..."),	NULL, folderview_new_folder_cb, 0, NULL},
 	{N_("/_Rename folder..."),	NULL, folderview_rename_folder_cb, 0, NULL},
+	{N_("/_Move folder..."),	NULL, folderview_move_folder_cb, 0, NULL},
 	{N_("/_Delete folder"),		NULL, folderview_delete_folder_cb, 0, NULL},
 	{N_("/---"),			NULL, NULL, 0, "<Separator>"},
 	{N_("/Empty _trash"),		NULL, folderview_empty_trash_cb, 0, NULL},
@@ -233,6 +247,7 @@ static GtkItemFactoryEntry folderview_imap_popup_entries[] =
 {
 	{N_("/Create _new folder..."),	NULL, folderview_new_folder_cb,    0, NULL},
 	{N_("/_Rename folder..."),	NULL, folderview_rename_folder_cb, 0, NULL},
+	{N_("/_Move folder..."),	NULL, folderview_move_folder_cb, 0, NULL},
 	{N_("/_Delete folder"),		NULL, folderview_delete_folder_cb, 0, NULL},
 	{N_("/---"),			NULL, NULL, 0, "<Separator>"},
 	{N_("/Empty _trash"),		NULL, folderview_empty_trash_cb, 0, NULL},
@@ -310,6 +325,7 @@ FolderView *folderview_create(void)
 	gtk_tree_view_set_search_column(GTK_TREE_VIEW(treeview),
 					COL_FOLDER_NAME);
 	gtk_tree_view_set_reorderable(GTK_TREE_VIEW(treeview), FALSE);
+	/* g_object_set(treeview, "fixed-height-mode", TRUE, NULL); */
 
 	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeview));
 	gtk_tree_selection_set_mode(selection, GTK_SELECTION_BROWSE);
@@ -352,7 +368,6 @@ FolderView *folderview_create(void)
 
 	gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), column);
 	gtk_tree_view_set_expander_column(GTK_TREE_VIEW(treeview), column);
-
 	g_signal_connect(G_OBJECT(column->button), "size-allocate",
 			 G_CALLBACK(folderview_col_resized), folderview);
 
@@ -361,33 +376,42 @@ FolderView *folderview_create(void)
 	column = gtk_tree_view_column_new_with_attributes
 		(_("New"), renderer, "text", COL_NEW, NULL);
 	gtk_tree_view_column_set_alignment(column, 1.0);
-	gtk_tree_view_column_set_sizing(column, GTK_TREE_VIEW_COLUMN_GROW_ONLY);
+	gtk_tree_view_column_set_sizing(column, GTK_TREE_VIEW_COLUMN_FIXED);
 	gtk_tree_view_column_set_fixed_width
 		(column, prefs_common.folder_col_new);
+	gtk_tree_view_column_set_min_width(column, 8);
 	gtk_tree_view_column_set_resizable(column, TRUE);
 	gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), column);
+	g_signal_connect(G_OBJECT(column->button), "size-allocate",
+			 G_CALLBACK(folderview_col_resized), folderview);
 
 	renderer = gtk_cell_renderer_text_new();
 	g_object_set(renderer, "xalign", 1.0, "ypad", 0, NULL);
 	column = gtk_tree_view_column_new_with_attributes
 		(_("Unread"), renderer, "text", COL_UNREAD, NULL);
 	gtk_tree_view_column_set_alignment(column, 1.0);
-	gtk_tree_view_column_set_sizing(column, GTK_TREE_VIEW_COLUMN_GROW_ONLY);
+	gtk_tree_view_column_set_sizing(column, GTK_TREE_VIEW_COLUMN_FIXED);
 	gtk_tree_view_column_set_fixed_width
 		(column, prefs_common.folder_col_unread);
+	gtk_tree_view_column_set_min_width(column, 8);
 	gtk_tree_view_column_set_resizable(column, TRUE);
 	gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), column);
+	g_signal_connect(G_OBJECT(column->button), "size-allocate",
+			 G_CALLBACK(folderview_col_resized), folderview);
 
 	renderer = gtk_cell_renderer_text_new();
 	g_object_set(renderer, "xalign", 1.0, "ypad", 0, NULL);
 	column = gtk_tree_view_column_new_with_attributes
 		(_("#"), renderer, "text", COL_TOTAL, NULL);
 	gtk_tree_view_column_set_alignment(column, 1.0);
-	gtk_tree_view_column_set_sizing(column, GTK_TREE_VIEW_COLUMN_GROW_ONLY);
+	gtk_tree_view_column_set_sizing(column, GTK_TREE_VIEW_COLUMN_FIXED);
 	gtk_tree_view_column_set_fixed_width
 		(column, prefs_common.folder_col_total);
+	gtk_tree_view_column_set_min_width(column, 8);
 	gtk_tree_view_column_set_resizable(column, TRUE);
 	gtk_tree_view_append_column(GTK_TREE_VIEW(treeview), column);
+	g_signal_connect(G_OBJECT(column->button), "size-allocate",
+			 G_CALLBACK(folderview_col_resized), folderview);
 
 	gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(store),
 					     COL_FOLDER_NAME,
@@ -438,7 +462,8 @@ FolderView *folderview_create(void)
 			 G_CALLBACK(folderview_popup_close), folderview);
 
         /* drop callback */
-	gtk_drag_dest_set(treeview, GTK_DEST_DEFAULT_ALL, summary_drag_types, 1,
+	gtk_drag_dest_set(treeview, GTK_DEST_DEFAULT_ALL,
+			  folderview_drag_types, 1,
 			  GDK_ACTION_MOVE | GDK_ACTION_COPY);
 	g_signal_connect(G_OBJECT(treeview), "drag-motion",
 			 G_CALLBACK(folderview_drag_motion_cb),
@@ -478,6 +503,7 @@ void folderview_init(FolderView *folderview)
 	stock_pixbuf_gdk(treeview, STOCK_PIXMAP_DIR_OPEN, &folderopen_pixbuf);
 	stock_pixbuf_gdk(treeview, STOCK_PIXMAP_DIR_NOSELECT,
 			 &foldernoselect_pixbuf);
+	stock_pixbuf_gdk(treeview, STOCK_PIXMAP_DRAFT, &draft_pixbuf);
 	stock_pixbuf_gdk(treeview, STOCK_PIXMAP_TRASH, &trash_pixbuf);
 }
 
@@ -533,26 +559,21 @@ void folderview_select(FolderView *folderview, FolderItem *item)
 static void folderview_select_row(FolderView *folderview, GtkTreeIter *iter)
 {
 	GtkTreeModel *model = GTK_TREE_MODEL(folderview->store);
-	GtkTreePath *path, *parent_path;
-	GtkTreeIter parent;
+	GtkTreePath *path;
 
 	g_return_if_fail(iter != NULL);
 
 	path = gtk_tree_model_get_path(model, iter);
 
-	if (gtk_tree_model_iter_parent(model, &parent, iter)) {
-		parent_path = gtk_tree_model_get_path(model, &parent);
-		gtk_tree_view_expand_to_path
-			(GTK_TREE_VIEW(folderview->treeview), parent_path);
-		gtk_tree_path_free(parent_path);
-	}
+	gtkut_tree_view_expand_parent_all(GTK_TREE_VIEW(folderview->treeview),
+					  iter);
 
 	folderview->open_folder = TRUE;
 	gtk_tree_view_set_cursor(GTK_TREE_VIEW(folderview->treeview), path,
 				 NULL, FALSE);
 	if (folderview->summaryview->folder_item &&
 	    folderview->summaryview->folder_item->total > 0)
-		gtk_widget_grab_focus(folderview->summaryview->ctree);
+		gtk_widget_grab_focus(folderview->summaryview->treeview);
 	else
 		gtk_widget_grab_focus(folderview->treeview);
 
@@ -568,6 +589,8 @@ static void folderview_select_row_ref(FolderView *folderview,
 	if (!row) return;
 
 	path = gtk_tree_row_reference_get_path(row);
+	if (!path)
+		return;
 	gtk_tree_model_get_iter(GTK_TREE_MODEL(folderview->store), &iter, path);
 	gtk_tree_path_free(path);
 
@@ -624,6 +647,8 @@ void folderview_select_next_unread(FolderView *folderview)
 		GtkTreePath *path;
 
 		path = gtk_tree_row_reference_get_path(folderview->opened);
+		if (!path)
+			return;
 		gtk_tree_model_get_iter(model, &iter, path);
 		gtk_tree_path_free(path);
 	} else {
@@ -653,6 +678,8 @@ FolderItem *folderview_get_selected_item(FolderView *folderview)
 		return NULL;
 
 	path = gtk_tree_row_reference_get_path(folderview->selected);
+	if (!path)
+		return NULL;
 	gtk_tree_model_get_iter(GTK_TREE_MODEL(folderview->store), &iter, path);
 	gtk_tree_path_free(path);
 	gtk_tree_model_get(GTK_TREE_MODEL(folderview->store), &iter,
@@ -661,6 +688,26 @@ FolderItem *folderview_get_selected_item(FolderView *folderview)
 	return item;
 }
 
+void folderview_set_opened_item(FolderView *folderview, FolderItem *item)
+{
+	GtkTreeModel *model = GTK_TREE_MODEL(folderview->store);
+	GtkTreeIter iter;
+	GtkTreePath *path;
+
+	gtk_tree_row_reference_free(folderview->opened);
+	folderview->opened = NULL;
+
+	if (!item)
+		return;
+
+	if (gtkut_tree_model_find_by_column_data
+		 (model, &iter, NULL, COL_FOLDER_ITEM, item)) {
+		path = gtk_tree_model_get_path(model, &iter);
+		folderview->opened = gtk_tree_row_reference_new(model, path);
+		gtk_tree_path_free(path);
+	}
+}
+
 void folderview_update_opened_msg_num(FolderView *folderview)
 {
 	GtkTreePath *path;
@@ -670,6 +717,8 @@ void folderview_update_opened_msg_num(FolderView *folderview)
 		return;
 
 	path = gtk_tree_row_reference_get_path(folderview->opened);
+	if (!path)
+		return;
 	gtk_tree_model_get_iter(GTK_TREE_MODEL(folderview->store), &iter, path);
 	gtk_tree_path_free(path);
 
@@ -677,7 +726,7 @@ void folderview_update_opened_msg_num(FolderView *folderview)
 }
 
 gboolean folderview_append_item(FolderView *folderview, GtkTreeIter *iter,
-				FolderItem *item)
+				FolderItem *item, gboolean expand_parent)
 {
 	FolderItem *parent_item;
 	GtkTreeModel *model = GTK_TREE_MODEL(folderview->store);
@@ -705,6 +754,15 @@ gboolean folderview_append_item(FolderView *folderview, GtkTreeIter *iter,
 		folderview_update_row(folderview, &child);
 		if (iter)
 			*iter = child;
+		if (expand_parent && iter_p) {
+			GtkTreePath *path;
+
+			path = gtk_tree_model_get_path(model, iter_p);
+			gtk_tree_view_expand_row
+				(GTK_TREE_VIEW(folderview->treeview),
+				 path, FALSE);
+			gtk_tree_path_free(path);
+		}
 		return TRUE;
 	}
 
@@ -803,6 +861,7 @@ static void folderview_rescan_tree(FolderView *folderview, Folder *folder)
 	window = label_window_create(_("Rebuilding folder tree..."));
 
 	summary_show(folderview->summaryview, NULL, FALSE);
+	GTK_EVENTS_FLUSH();
 
 	folder_set_ui_func(folder, folderview_scan_tree_func, NULL);
 	if (folder->klass->scan_tree(folder) < 0)
@@ -819,51 +878,49 @@ static void folderview_rescan_tree(FolderView *folderview, Folder *folder)
 
 void folderview_check_new(Folder *folder)
 {
-	GList *list;
 	FolderItem *item;
 	FolderView *folderview;
 	GtkTreeModel *model;
 	GtkTreeIter iter;
 	gboolean valid;
 
-	for (list = folderview_list; list != NULL; list = list->next) {
-		folderview = (FolderView *)list->data;
-		model = GTK_TREE_MODEL(folderview->store);
+	folderview = (FolderView *)folderview_list->data;
+	model = GTK_TREE_MODEL(folderview->store);
 
-		if (folder && !FOLDER_IS_LOCAL(folder)) {
-			if (!main_window_toggle_online_if_offline
-				(folderview->mainwin))
-				return;
-		}
-
-		inc_lock();
-		main_window_lock(folderview->mainwin);
-		gtk_widget_set_sensitive(folderview->treeview, FALSE);
-
-		for (valid = gtk_tree_model_get_iter_first(model, &iter);
-		     valid; valid = gtkut_tree_model_next(model, &iter)) {
-			item = NULL;
-			gtk_tree_model_get(model, &iter,
-					   COL_FOLDER_ITEM, &item, -1);
-			if (!item || !item->path || !item->folder) continue;
-			if (item->no_select) continue;
-			if (folder && folder != item->folder) continue;
-			if (!folder && !FOLDER_IS_LOCAL(item->folder)) continue;
-
-			folderview_scan_tree_func(item->folder, item, NULL);
-			if (folder_item_scan(item) < 0) {
-				if (folder && !FOLDER_IS_LOCAL(folder))
-					break;
-			}
-			folderview_update_row(folderview, &iter);
-		}
-
-		gtk_widget_set_sensitive(folderview->treeview, TRUE);
-		main_window_unlock(folderview->mainwin);
-		inc_unlock();
-		statusbar_pop_all();
+	if (folder && !FOLDER_IS_LOCAL(folder)) {
+		if (!main_window_toggle_online_if_offline
+			(folderview->mainwin))
+			return;
 	}
 
+	inc_lock();
+	main_window_lock(folderview->mainwin);
+	gtk_widget_set_sensitive(folderview->treeview, FALSE);
+	GTK_EVENTS_FLUSH();
+
+	for (valid = gtk_tree_model_get_iter_first(model, &iter);
+	     valid; valid = gtkut_tree_model_next(model, &iter)) {
+		item = NULL;
+		gtk_tree_model_get(model, &iter,
+				   COL_FOLDER_ITEM, &item, -1);
+		if (!item || !item->path || !item->folder) continue;
+		if (item->no_select) continue;
+		if (folder && folder != item->folder) continue;
+		if (!folder && !FOLDER_IS_LOCAL(item->folder)) continue;
+
+		folderview_scan_tree_func(item->folder, item, NULL);
+		if (folder_item_scan(item) < 0) {
+			if (folder && !FOLDER_IS_LOCAL(folder))
+				break;
+		}
+		folderview_update_row(folderview, &iter);
+	}
+
+	gtk_widget_set_sensitive(folderview->treeview, TRUE);
+	main_window_unlock(folderview->mainwin);
+	inc_unlock();
+	statusbar_pop_all();
+
 	folder_write_list();
 }
 
@@ -1031,8 +1088,7 @@ static void folderview_update_row(FolderView *folderview, GtkTreeIter *iter)
 				item->name);
 		break;
 	case F_DRAFT:
-		pixbuf = folder_pixbuf;
-		open_pixbuf = folderopen_pixbuf;
+		pixbuf = open_pixbuf = draft_pixbuf;
 		name = g_strdup(FOLDER_IS_LOCAL(item->folder) &&
 				!strcmp2(item->name, DRAFT_DIR) ? _("Drafts") :
 				item->name);
@@ -1203,32 +1259,37 @@ void folderview_update_all_updated(gboolean update_summary)
 	}
 }
 
-static void folderview_insert_item_recursive(FolderView *folderview,
-					     FolderItem *item)
+static gboolean folderview_insert_item_recursive(FolderView *folderview,
+						 FolderItem *item,
+						 GtkTreeIter *iter)
 {
 	GNode *node;
-	GtkTreeIter iter;
+	GtkTreeIter iter_;
 	gboolean valid;
 
-	g_return_if_fail(item != NULL);
+	g_return_val_if_fail(item != NULL, FALSE);
 
-	valid = folderview_append_item(folderview, &iter, item);
-	g_return_if_fail(valid == TRUE);
+	valid = folderview_append_item(folderview, &iter_, item, FALSE);
+	g_return_val_if_fail(valid == TRUE, FALSE);
 
 	for (node = item->node->children; node != NULL; node = node->next) {
 		FolderItem *child_item = FOLDER_ITEM(node->data);
-		folderview_insert_item_recursive(folderview, child_item);
+		folderview_insert_item_recursive(folderview, child_item, NULL);
 	}
 
 	if (item->node->children && !item->collapsed) {
 		GtkTreePath *path;
 
 		path = gtk_tree_model_get_path
-			(GTK_TREE_MODEL(folderview->store), &iter);
+			(GTK_TREE_MODEL(folderview->store), &iter_);
 		gtk_tree_view_expand_row(GTK_TREE_VIEW(folderview->treeview),
 					 path, FALSE);
 		gtk_tree_path_free(path);
 	}
+
+	if (iter)
+		*iter = iter_;
+	return TRUE;
 }
 
 static void folderview_append_folder(FolderView *folderview, Folder *folder)
@@ -1236,7 +1297,7 @@ static void folderview_append_folder(FolderView *folderview, Folder *folder)
 	g_return_if_fail(folder != NULL);
 
 	folderview_insert_item_recursive
-		(folderview, FOLDER_ITEM(folder->node->data));
+		(folderview, FOLDER_ITEM(folder->node->data), NULL);
 }
 
 void folderview_new_folder(FolderView *folderview)
@@ -1249,19 +1310,10 @@ void folderview_new_folder(FolderView *folderview)
 
 	g_return_if_fail(item->folder != NULL);
 
-	switch (FOLDER_TYPE(item->folder)) {
-	case F_MH:
-	case F_MBOX:
-	case F_MAILDIR:
-	case F_IMAP:
+	if (item->folder->klass->create_folder)
 		folderview_new_folder_cb(folderview, 0, NULL);
-		break;
-	case F_NEWS:
+	else if (FOLDER_TYPE(item->folder) == F_NEWS)
 		folderview_new_news_group_cb(folderview, 0, NULL);
-		break;
-	default:
-		break;
-	}
 }
 
 void folderview_rename_folder(FolderView *folderview)
@@ -1277,17 +1329,25 @@ void folderview_rename_folder(FolderView *folderview)
 	if (!item->path) return;
 	if (item->stype != F_NORMAL) return;
 
-	switch (FOLDER_TYPE(item->folder)) {
-	case F_MH:
-	case F_MBOX:
-	case F_MAILDIR:
-	case F_IMAP:
+	if (item->folder->klass->rename_folder)
 		folderview_rename_folder_cb(folderview, 0, NULL);
-		break;
-	case F_NEWS:
-	default:
-		break;
-	}
+}
+
+void folderview_move_folder(FolderView *folderview)
+{
+	FolderItem *item;
+
+	item = folderview_get_selected_item(folderview);
+	if (!item)
+		return;
+
+	g_return_if_fail(item->folder != NULL);
+
+	if (!item->path) return;
+	if (item->stype != F_NORMAL) return;
+
+	if (item->folder->klass->move_folder)
+		folderview_move_folder_cb(folderview, 0, NULL);
 }
 
 void folderview_delete_folder(FolderView *folderview)
@@ -1303,19 +1363,10 @@ void folderview_delete_folder(FolderView *folderview)
 	if (!item->path) return;
 	if (item->stype != F_NORMAL) return;
 
-	switch (FOLDER_TYPE(item->folder)) {
-	case F_MH:
-	case F_MBOX:
-	case F_MAILDIR:
-	case F_IMAP:
+	if (item->folder->klass->remove_folder)
 		folderview_delete_folder_cb(folderview, 0, NULL);
-		break;
-	case F_NEWS:
+	else if (FOLDER_TYPE(item->folder) == F_NEWS)
 		folderview_rm_news_group_cb(folderview, 0, NULL);
-		break;
-	default:
-		break;
-	}
 }
 
 void folderview_check_new_selected(FolderView *folderview)
@@ -1384,6 +1435,7 @@ static gboolean folderview_menu_popup(FolderView *folderview,
 	GtkTreeIter iter;
 	gboolean new_folder      = FALSE;
 	gboolean rename_folder   = FALSE;
+	gboolean move_folder     = FALSE;
 	gboolean delete_folder   = FALSE;
 	gboolean empty_trash     = FALSE;
 	gboolean download_msg    = FALSE;
@@ -1422,9 +1474,11 @@ static gboolean folderview_menu_popup(FolderView *folderview,
 		if (FOLDER_IS_LOCAL(folder) || FOLDER_TYPE(folder) == F_IMAP) {
 			if (item->parent == NULL)
 				update_tree = rescan_tree = TRUE;
-			else if (item->stype == F_NORMAL)
+			else if (item->stype == F_NORMAL) {
 				rename_folder = delete_folder = TRUE;
-			else if (item->stype == F_TRASH)
+				if (folder->klass->move_folder)
+					move_folder = TRUE;
+			} else if (item->stype == F_TRASH)
 				empty_trash = TRUE;
 		} else if (FOLDER_TYPE(folder) == F_NEWS) {
 			if (item->parent != NULL)
@@ -1445,6 +1499,7 @@ static gboolean folderview_menu_popup(FolderView *folderview,
 		menu_set_insensitive_all(GTK_MENU_SHELL(popup));
 		SET_SENS(mail_factory, "/Create new folder...", new_folder);
 		SET_SENS(mail_factory, "/Rename folder...", rename_folder);
+		SET_SENS(mail_factory, "/Move folder...", move_folder);
 		SET_SENS(mail_factory, "/Delete folder", delete_folder);
 		SET_SENS(mail_factory, "/Empty trash", empty_trash);
 		SET_SENS(mail_factory, "/Check for new messages", update_tree);
@@ -1456,6 +1511,7 @@ static gboolean folderview_menu_popup(FolderView *folderview,
 		menu_set_insensitive_all(GTK_MENU_SHELL(popup));
 		SET_SENS(imap_factory, "/Create new folder...", new_folder);
 		SET_SENS(imap_factory, "/Rename folder...", rename_folder);
+		SET_SENS(imap_factory, "/Move folder...", move_folder);
 		SET_SENS(imap_factory, "/Delete folder", delete_folder);
 		SET_SENS(imap_factory, "/Empty trash", empty_trash);
 		SET_SENS(imap_factory, "/Download", download_msg);
@@ -1500,15 +1556,24 @@ static gboolean folderview_button_pressed(GtkWidget *widget,
 					   &path, NULL, NULL, NULL))
 		return TRUE;
 
+	if (folderview->selection_locked)
+		return TRUE;
+
 	if (event->button == 1 || event->button == 2) {
 		folderview->open_folder = TRUE;
 	} else if (event->button == 3) {
-		gtk_tree_view_set_cursor(treeview, path, NULL, FALSE);
+		if (folderview->selected) {
+			folderview->prev_selected =
+				gtk_tree_row_reference_copy
+					(folderview->selected);
+		}
+		gtk_tree_selection_select_path(folderview->selection, path);
 		folderview_menu_popup(folderview, event);
+		gtk_tree_path_free(path);
+		return TRUE;
 	}
 
 	gtk_tree_path_free(path);
-
 	return FALSE;
 }
 
@@ -1527,6 +1592,9 @@ static gboolean folderview_key_pressed(GtkWidget *widget, GdkEventKey *event,
 
 	if (!event) return FALSE;
 
+	if (folderview->selection_locked)
+		return TRUE;
+
 	switch (event->keyval) {
 	case GDK_Return:
 		if (folderview->selected) {
@@ -1542,7 +1610,8 @@ static gboolean folderview_key_pressed(GtkWidget *widget, GdkEventKey *event,
 					(folderview->opened);
 			selected = gtk_tree_row_reference_get_path
 				(folderview->selected);
-			if (gtk_tree_path_compare(opened, selected) == 0 &&
+			if (opened && selected &&
+			    gtk_tree_path_compare(opened, selected) == 0 &&
 			    (!folderview->summaryview->folder_item ||
 			     folderview->summaryview->folder_item->total == 0))
 				folderview_select_next_unread(folderview);
@@ -1579,8 +1648,13 @@ static void folderview_selection_changed(GtkTreeSelection *selection,
 	GtkTreePath *path;
 	gboolean opened;
 
-	if (!gtk_tree_selection_get_selected(selection, NULL, &iter))
+	if (!gtk_tree_selection_get_selected(selection, NULL, &iter)) {
+		if (folderview->selected) {
+			gtk_tree_row_reference_free(folderview->selected);
+			folderview->selected = NULL;
+		}
 		return;
+	}
 
 	path = gtk_tree_model_get_path(model, &iter);
 
@@ -1613,7 +1687,7 @@ static void folderview_selection_changed(GtkTreeSelection *selection,
 		GtkTreePath *open_path = NULL;
 
 		open_path = gtk_tree_row_reference_get_path(folderview->opened);
-		if (gtk_tree_path_compare(open_path, path) == 0) {
+		if (open_path && gtk_tree_path_compare(open_path, path) == 0) {
 			gtk_tree_path_free(open_path);
 			gtk_tree_path_free(path);
 			return;
@@ -1621,6 +1695,9 @@ static void folderview_selection_changed(GtkTreeSelection *selection,
 		gtk_tree_path_free(open_path);
 	}
 
+	folderview->selection_locked = TRUE;
+
+	GTK_EVENTS_FLUSH();
 	opened = summary_show(folderview->summaryview, item, FALSE);
 
 	if (opened) {
@@ -1640,6 +1717,8 @@ static void folderview_selection_changed(GtkTreeSelection *selection,
 		folderview_select_row_ref(folderview, folderview->opened);
 
 	gtk_tree_path_free(path);
+
+	folderview->selection_locked = FALSE;
 }
 
 static void folderview_row_expanded(GtkTreeView *treeview, GtkTreeIter *iter,
@@ -1698,18 +1777,47 @@ static void folderview_popup_close(GtkMenuShell *menu_shell,
 {
 	GtkTreePath *path;
 
-	if (!folderview->opened) return;
+	if (!folderview->prev_selected) return;
 
-	path = gtk_tree_row_reference_get_path(folderview->opened);
-	gtk_tree_view_set_cursor(GTK_TREE_VIEW(folderview->treeview), path,
-				 NULL, FALSE);
+	path = gtk_tree_row_reference_get_path(folderview->prev_selected);
+	gtk_tree_row_reference_free(folderview->prev_selected);
+	folderview->prev_selected = NULL;
+	if (!path)
+		return;
+	gtk_tree_selection_select_path(folderview->selection, path);
 	gtk_tree_path_free(path);
 }
 
 static void folderview_col_resized(GtkWidget *widget, GtkAllocation *allocation,
 				   FolderView *folderview)
 {
-	prefs_common.folder_col_folder = allocation->width;
+	GtkTreeViewColumn *column;
+	gint type;
+	gint width = allocation->width;
+
+	for (type = 0; type <= COL_TOTAL; type++) {
+		column = gtk_tree_view_get_column
+			(GTK_TREE_VIEW(folderview->treeview), type);
+		if (column && column->button == widget) {
+			switch (type) {
+			case COL_FOLDER_NAME:
+				prefs_common.folder_col_folder = width;
+				break;
+			case COL_NEW:
+				prefs_common.folder_col_new = width;
+				break;
+			case COL_UNREAD:
+				prefs_common.folder_col_unread = width;
+				break;
+			case COL_TOTAL:
+				prefs_common.folder_col_total = width;
+				break;
+			default:
+				break;
+			}
+			break;
+		}
+	}
 }
 
 static void folderview_download_func(Folder *folder, FolderItem *item,
@@ -1844,7 +1952,7 @@ static void folderview_new_folder_cb(FolderView *folderview, guint action,
 		return;
 	}
 
-	folderview_append_item(folderview, NULL, new_item);
+	folderview_append_item(folderview, NULL, new_item, TRUE);
 	folder_write_list();
 }
 
@@ -1859,7 +1967,7 @@ static void folderview_rename_folder_cb(FolderView *folderview, guint action,
 	gchar *old_id;
 	gchar *new_id;
 	GtkTreePath *sel_path;
-	GtkTreePath *open_path;
+	GtkTreePath *open_path = NULL;
 	GtkTreeIter iter;
 
 	item = folderview_get_selected_item(folderview);
@@ -1908,12 +2016,16 @@ static void folderview_rename_folder_cb(FolderView *folderview, guint action,
 	g_free(new_id);
 
 	sel_path = gtk_tree_row_reference_get_path(folderview->selected);
-	open_path = gtk_tree_row_reference_get_path(folderview->opened);
-	gtk_tree_model_get_iter(GTK_TREE_MODEL(folderview->store),
-				&iter, sel_path);
-	folderview_update_row(folderview, &iter);
-	if (gtk_tree_path_compare(open_path, sel_path) == 0 ||
-	    gtk_tree_path_is_ancestor(sel_path, open_path)) {
+	if (folderview->opened)
+		open_path = gtk_tree_row_reference_get_path(folderview->opened);
+	if (sel_path) {
+		gtk_tree_model_get_iter(GTK_TREE_MODEL(folderview->store),
+					&iter, sel_path);
+		folderview_update_row(folderview, &iter);
+	}
+	if (sel_path && open_path &&
+	    (gtk_tree_path_compare(open_path, sel_path) == 0 ||
+	     gtk_tree_path_is_ancestor(sel_path, open_path))) {
 		GtkTreeRowReference *row;
 
 		row = gtk_tree_row_reference_copy(folderview->opened);
@@ -1927,6 +2039,76 @@ static void folderview_rename_folder_cb(FolderView *folderview, guint action,
 	folder_write_list();
 }
 
+static void folderview_move_folder_cb(FolderView *folderview, guint action,
+				      GtkWidget *widget)
+{
+	FolderItem *item;
+	FolderItem *new_parent;
+	GtkTreePath *sel_path;
+	GtkTreePath *open_path = NULL;
+	GtkTreeIter iter;
+	gchar *old_path, *old_id, *new_id;
+
+	item = folderview_get_selected_item(folderview);
+	if (!item)
+		return;
+
+	g_return_if_fail(item->path != NULL);
+	g_return_if_fail(item->folder != NULL);
+
+	new_parent = foldersel_folder_sel(item->folder, FOLDER_SEL_MOVE_FOLDER,
+					  NULL);
+	if (!new_parent)
+		return;
+	if (new_parent->folder != item->folder)
+		return;
+	if (new_parent == item->parent)
+		return;
+
+	old_path = g_strdup(item->path);
+	old_id = folder_item_get_identifier(item);
+
+	if (item->folder->klass->move_folder
+		(item->folder, item, new_parent) < 0) {
+		alertpanel_error(_("Can't move the folder `%s'."), item->name);
+		g_free(old_id);
+		g_free(old_path);
+		return;
+	}
+
+	if (folder_get_default_folder() == item->folder)
+		prefs_filter_rename_path(old_path, item->path);
+	new_id = folder_item_get_identifier(item);
+	prefs_filter_rename_path(old_id, new_id);
+	g_free(new_id);
+	g_free(old_id);
+	g_free(old_path);
+
+	sel_path = gtk_tree_row_reference_get_path(folderview->selected);
+	if (folderview->opened)
+		open_path = gtk_tree_row_reference_get_path(folderview->opened);
+	g_return_if_fail(sel_path != NULL);
+	gtk_tree_model_get_iter(GTK_TREE_MODEL(folderview->store), &iter,
+				sel_path);
+	if (sel_path && open_path &&
+	    (gtk_tree_path_compare(open_path, sel_path) == 0 ||
+	     gtk_tree_path_is_ancestor(sel_path, open_path))) {
+		summary_clear_all(folderview->summaryview);
+		gtk_tree_row_reference_free(folderview->opened);
+		folderview->opened = NULL;
+	}
+	gtk_tree_path_free(open_path);
+	gtk_tree_path_free(sel_path);
+
+	gtk_tree_store_remove(folderview->store, &iter);
+	if (folderview_insert_item_recursive(folderview, item, &iter)) {
+		gtkut_tree_view_expand_parent_all
+			(GTK_TREE_VIEW(folderview->treeview), &iter);
+	}
+
+	folder_write_list();
+}
+
 static void folderview_delete_folder_cb(FolderView *folderview, guint action,
 					GtkWidget *widget)
 {
@@ -1936,7 +2118,7 @@ static void folderview_delete_folder_cb(FolderView *folderview, guint action,
 	AlertValue avalue;
 	gchar *old_path;
 	gchar *old_id;
-	GtkTreePath *sel_path, *open_path;
+	GtkTreePath *sel_path, *open_path = NULL;
 	GtkTreeIter iter;
 
 	item = folderview_get_selected_item(folderview);
@@ -1951,10 +2133,12 @@ static void folderview_delete_folder_cb(FolderView *folderview, guint action,
 	name = trim_string(item->name, 32);
 	AUTORELEASE_STR(name, {g_free(name); return;});
 	message = g_strdup_printf
-		(_("All folder(s) and message(s) under `%s' will be deleted.\n"
+		(_("All folders and messages under `%s' will be permanently deleted.\n"
+		   "Recovery will not be possible.\n\n"
 		   "Do you really want to delete?"), name);
-	avalue = alertpanel(_("Delete folder"), message,
-			    _("Yes"), _("+No"), NULL);
+	avalue = alertpanel_full(_("Delete folder"), message,
+				 ALERT_WARNING, G_ALERTALTERNATE, FALSE,
+				 GTK_STOCK_YES, GTK_STOCK_NO, NULL);
 	g_free(message);
 	if (avalue != G_ALERTDEFAULT) return;
 
@@ -1962,11 +2146,14 @@ static void folderview_delete_folder_cb(FolderView *folderview, guint action,
 	old_id = folder_item_get_identifier(item);
 
 	sel_path = gtk_tree_row_reference_get_path(folderview->selected);
-	open_path = gtk_tree_row_reference_get_path(folderview->opened);
+	g_return_if_fail(sel_path != NULL);
+	if (folderview->opened)
+		open_path = gtk_tree_row_reference_get_path(folderview->opened);
 	gtk_tree_model_get_iter(GTK_TREE_MODEL(folderview->store), &iter,
 				sel_path);
-	if (gtk_tree_path_compare(open_path, sel_path) == 0 ||
-	    gtk_tree_path_is_ancestor(sel_path, open_path)) {
+	if (sel_path && open_path &&
+	    (gtk_tree_path_compare(open_path, sel_path) == 0 ||
+	     gtk_tree_path_is_ancestor(sel_path, open_path))) {
 		summary_clear_all(folderview->summaryview);
 		gtk_tree_row_reference_free(folderview->opened);
 		folderview->opened = NULL;
@@ -2019,7 +2206,8 @@ static void folderview_empty_trash_cb(FolderView *folderview, guint action,
 
 	sel_path = gtk_tree_row_reference_get_path(folderview->selected);
 	open_path = gtk_tree_row_reference_get_path(folderview->opened);
-	if (gtk_tree_path_compare(open_path, sel_path) == 0)
+	if (open_path && sel_path &&
+	    gtk_tree_path_compare(open_path, sel_path) == 0)
 		gtk_widget_grab_focus(folderview->treeview);
 	gtk_tree_path_free(open_path);
 	gtk_tree_path_free(sel_path);
@@ -2046,8 +2234,9 @@ static void folderview_remove_mailbox_cb(FolderView *folderview, guint action,
 	message = g_strdup_printf
 		(_("Really remove the mailbox `%s' ?\n"
 		   "(The messages are NOT deleted from the disk)"), name);
-	avalue = alertpanel(_("Remove mailbox"), message,
-			    _("Yes"), _("+No"), NULL);
+	avalue = alertpanel_full(_("Remove mailbox"), message,
+				 ALERT_WARNING, G_ALERTALTERNATE, FALSE,
+				 GTK_STOCK_YES, GTK_STOCK_NO, NULL);
 	g_free(message);
 	g_free(name);
 	if (avalue != G_ALERTDEFAULT) return;
@@ -2061,10 +2250,12 @@ static void folderview_remove_mailbox_cb(FolderView *folderview, guint action,
 	folder_destroy(item->folder);
 
 	sel_path = gtk_tree_row_reference_get_path(folderview->selected);
-	gtk_tree_model_get_iter(GTK_TREE_MODEL(folderview->store), &iter,
-				sel_path);
-	gtk_tree_path_free(sel_path);
-	gtk_tree_store_remove(folderview->store, &iter);
+	if (sel_path) {
+		gtk_tree_model_get_iter(GTK_TREE_MODEL(folderview->store),
+					&iter, sel_path);
+		gtk_tree_path_free(sel_path);
+		gtk_tree_store_remove(folderview->store, &iter);
+	}
 
 	folder_write_list();
 }
@@ -2090,8 +2281,9 @@ static void folderview_rm_imap_server_cb(FolderView *folderview, guint action,
 
 	name = trim_string(item->folder->name, 32);
 	message = g_strdup_printf(_("Really delete IMAP4 account `%s'?"), name);
-	avalue = alertpanel(_("Delete IMAP4 account"), message,
-			    _("Yes"), _("+No"), NULL);
+	avalue = alertpanel_full(_("Delete IMAP4 account"), message,
+				 ALERT_WARNING, G_ALERTALTERNATE, FALSE,
+				 GTK_STOCK_YES, GTK_STOCK_NO, NULL);
 	g_free(message);
 	g_free(name);
 
@@ -2109,10 +2301,12 @@ static void folderview_rm_imap_server_cb(FolderView *folderview, guint action,
 	account_destroy(account);
 
 	sel_path = gtk_tree_row_reference_get_path(folderview->selected);
-	gtk_tree_model_get_iter(GTK_TREE_MODEL(folderview->store), &iter,
-				sel_path);
-	gtk_tree_path_free(sel_path);
-	gtk_tree_store_remove(folderview->store, &iter);
+	if (sel_path) {
+		gtk_tree_model_get_iter(GTK_TREE_MODEL(folderview->store),
+					&iter, sel_path);
+		gtk_tree_path_free(sel_path);
+		gtk_tree_store_remove(folderview->store, &iter);
+	}
 
 	account_set_menu();
 	main_window_reflect_prefs_all();
@@ -2143,6 +2337,7 @@ static void folderview_new_news_group_cb(FolderView *folderview, guint action,
 	g_return_if_fail(folder->account != NULL);
 
 	server_path = gtk_tree_row_reference_get_path(folderview->selected);
+	g_return_if_fail(server_path != NULL);
 	gtk_tree_model_get_iter(model, &iter, server_path);
 	gtk_tree_path_free(server_path);
 
@@ -2160,7 +2355,8 @@ static void folderview_new_news_group_cb(FolderView *folderview, guint action,
 
 		item = FOLDER_ITEM(gnode->data);
 		if (g_slist_find_custom(new_subscr, item->path,
-					(GCompareFunc)g_strcasecmp) != NULL) {
+					(GCompareFunc)g_ascii_strcasecmp)
+		    != NULL) {
 			gnode = next;
 			continue;
 		}
@@ -2192,7 +2388,7 @@ static void folderview_new_news_group_cb(FolderView *folderview, guint action,
 
 		newitem = folder_item_new(name, name);
 		folder_item_append(rootitem, newitem);
-		folderview_append_item(folderview, NULL, newitem);
+		folderview_append_item(folderview, NULL, newitem, TRUE);
 	}
 
 	if (new_subscr) {
@@ -2215,7 +2411,7 @@ static void folderview_rm_news_group_cb(FolderView *folderview, guint action,
 	gchar *name;
 	gchar *message;
 	AlertValue avalue;
-	GtkTreePath *sel_path, *open_path;
+	GtkTreePath *sel_path, *open_path = NULL;
 	GtkTreeIter iter;
 
 	item = folderview_get_selected_item(folderview);
@@ -2228,17 +2424,21 @@ static void folderview_rm_news_group_cb(FolderView *folderview, guint action,
 
 	name = trim_string_before(item->path, 32);
 	message = g_strdup_printf(_("Really delete newsgroup `%s'?"), name);
-	avalue = alertpanel(_("Delete newsgroup"), message,
-			    _("Yes"), _("+No"), NULL);
+	avalue = alertpanel_full(_("Delete newsgroup"), message,
+				 ALERT_WARNING, G_ALERTALTERNATE, FALSE,
+				 GTK_STOCK_YES, GTK_STOCK_NO, NULL);
 	g_free(message);
 	g_free(name);
 	if (avalue != G_ALERTDEFAULT) return;
 
 	sel_path = gtk_tree_row_reference_get_path(folderview->selected);
-	open_path = gtk_tree_row_reference_get_path(folderview->opened);
+	g_return_if_fail(sel_path != NULL);
+	if (folderview->opened)
+		open_path = gtk_tree_row_reference_get_path(folderview->opened);
 	gtk_tree_model_get_iter(GTK_TREE_MODEL(folderview->store), &iter,
 				sel_path);
-	if (gtk_tree_path_compare(open_path, sel_path) == 0) {
+	if (open_path && sel_path &&
+	    gtk_tree_path_compare(open_path, sel_path) == 0) {
 		summary_clear_all(folderview->summaryview);
 		gtk_tree_row_reference_free(folderview->opened);
 		folderview->opened = NULL;
@@ -2272,8 +2472,9 @@ static void folderview_rm_news_server_cb(FolderView *folderview, guint action,
 
 	name = trim_string(item->folder->name, 32);
 	message = g_strdup_printf(_("Really delete news account `%s'?"), name);
-	avalue = alertpanel(_("Delete news account"), message,
-			    _("Yes"), _("+No"), NULL);
+	avalue = alertpanel_full(_("Delete news account"), message,
+				 ALERT_WARNING, G_ALERTALTERNATE, FALSE,
+				 GTK_STOCK_YES, GTK_STOCK_NO, NULL);
 	g_free(message);
 	g_free(name);
 
@@ -2291,10 +2492,12 @@ static void folderview_rm_news_server_cb(FolderView *folderview, guint action,
 	account_destroy(account);
 
 	sel_path = gtk_tree_row_reference_get_path(folderview->selected);
-	gtk_tree_model_get_iter(GTK_TREE_MODEL(folderview->store), &iter,
-				sel_path);
-	gtk_tree_path_free(sel_path);
-	gtk_tree_store_remove(folderview->store, &iter);
+	if (sel_path) {
+		gtk_tree_model_get_iter(GTK_TREE_MODEL(folderview->store),
+					&iter, sel_path);
+		gtk_tree_path_free(sel_path);
+		gtk_tree_store_remove(folderview->store, &iter);
+	}
 
 	account_set_menu();
 	main_window_reflect_prefs_all();
@@ -2378,7 +2581,7 @@ static gboolean folderview_drag_motion_cb(GtkWidget      *widget,
 	GtkTreeModel *model = GTK_TREE_MODEL(folderview->store);
 	GtkTreePath *path = NULL, *prev_path = NULL;
 	GtkTreeIter iter;
-	FolderItem *item = NULL, *src_item;
+	FolderItem *item = NULL, *src_item = NULL;
 	gboolean acceptable = FALSE;
 
 	if (gtk_tree_view_get_dest_row_at_pos
@@ -2414,7 +2617,8 @@ static gboolean folderview_drag_motion_cb(GtkWidget      *widget,
 	}
 
 	if (acceptable) {
-		if ((context->actions & GDK_ACTION_MOVE) != 0)
+		if ((context->actions & GDK_ACTION_MOVE) != 0 &&
+		    FOLDER_ITEM_CAN_ADD(src_item))
 			gdk_drag_status(context, GDK_ACTION_MOVE, time);
 		else if ((context->actions & GDK_ACTION_COPY) != 0)
 			gdk_drag_status(context, GDK_ACTION_COPY, time);
@@ -2469,12 +2673,14 @@ static void folderview_drag_received_cb(GtkWidget        *widget,
 	src_item = folderview->summaryview->folder_item;
 
 	if (FOLDER_ITEM_CAN_ADD(item) && src_item && src_item != item) {
-		if ((context->actions & GDK_ACTION_MOVE) != 0) {
+		if ((context->actions & GDK_ACTION_MOVE) != 0 &&
+		    FOLDER_ITEM_CAN_ADD(src_item)) {
 			summary_move_selected_to(folderview->summaryview, item);
-			gtk_drag_finish(context, TRUE, TRUE, time);
+			context->action = 0;
+			gtk_drag_finish(context, TRUE, FALSE, time);
 		} else if ((context->actions & GDK_ACTION_COPY) != 0) {
 			summary_copy_selected_to(folderview->summaryview, item);
-			gtk_drag_finish(context, TRUE, TRUE, time);
+			gtk_drag_finish(context, TRUE, FALSE, time);
 		} else
 			gtk_drag_finish(context, FALSE, FALSE, time);
 	} else
