@@ -541,14 +541,14 @@ static void html_get_parenthesis(HTMLParser *parser, gchar *buf, gint len)
 		parser->bufp = p + 3;
 		return;
 	}
-	if (!g_strncasecmp(parser->bufp, "<style", 6)) {
+	if (!g_ascii_strncasecmp(parser->bufp, "<style", 6)) {
 		parser->bufp += 6;
 		while ((p = strcasestr(parser->bufp, "</style>")) == NULL)
 			if (html_read_line(parser) == HTML_EOF) return;
 		parser->bufp = p + 8;
 		return;
 	}
-	if (!g_strncasecmp(parser->bufp, "<script", 7)) {
+	if (!g_ascii_strncasecmp(parser->bufp, "<script", 7)) {
 		parser->bufp += 7;
 		while ((p = strcasestr(parser->bufp, "</script>")) == NULL)
 			if (html_read_line(parser) == HTML_EOF) return;
