@@ -84,6 +84,8 @@ static Folder	*imap_folder_new	(const gchar	*name,
 static void	 imap_folder_destroy	(Folder		*folder);
 
 static Session *imap_session_new	(PrefsAccount	*account);
+static gint imap_session_connect	(IMAPSession	*session);
+static gint imap_session_reconnect	(IMAPSession	*session);
 static void imap_session_destroy	(Session	*session);
 /* static void imap_session_destroy_all	(void); */
 
@@ -155,6 +157,9 @@ static FolderItem *imap_create_folder	(Folder		*folder,
 static gint imap_rename_folder		(Folder		*folder,
 					 FolderItem	*item,
 					 const gchar	*name);
+static gint imap_move_folder		(Folder		*folder,
+					 FolderItem	*item,
+					 FolderItem	*new_parent);
 static gint imap_remove_folder		(Folder		*folder,
 					 FolderItem	*item);
 
@@ -389,6 +394,7 @@ static FolderClass imap_class =
 
 	imap_create_folder,
 	imap_rename_folder,
+	imap_move_folder,
 	imap_remove_folder
 };
 
@@ -455,11 +461,14 @@ static IMAPSession *imap_session_get(Folder *folder)
 		log_warning(_("IMAP4 connection to %s has been"
 			      " disconnected. Reconnecting...\n"),
 			    folder->account->recv_server);
-		session_destroy(rfolder->session);
-		rfolder->session = imap_session_new(folder->account);
-		if (rfolder->session)
+		if (imap_session_reconnect(IMAP_SESSION(rfolder->session))
+		    == IMAP_SUCCESS)
 			imap_parse_namespace(IMAP_SESSION(rfolder->session),
 					     IMAP_FOLDER(folder));
+		else {
+			session_destroy(rfolder->session);
+			rfolder->session = NULL;
+		}
 	}
 
 	return IMAP_SESSION(rfolder->session);
@@ -490,12 +499,32 @@ static gint imap_greeting(IMAPSession *session)
 static gint imap_auth(IMAPSession *session, const gchar *user,
 		      const gchar *pass, IMAPAuthType type)
 {
-	gint ok;
+	gboolean nologin;
+	gint ok = IMAP_AUTHFAIL;
 
-	if (type == 0 || type == IMAP_AUTH_LOGIN)
-		ok = imap_cmd_login(session, user, pass);
-	else
+	nologin = imap_has_capability(session, "LOGINDISABLED");
+
+	switch (type) {
+	case 0:
+		if (imap_has_capability(session, "AUTH=CRAM-MD5"))
+			ok = imap_cmd_authenticate(session, user, pass, type);
+		else if (nologin)
+			log_print(_("IMAP4 server disables LOGIN.\n"));
+		else
+			ok = imap_cmd_login(session, user, pass);
+		break;
+	case IMAP_AUTH_LOGIN:
+		if (nologin)
+			log_warning(_("IMAP4 server disables LOGIN.\n"));
+		else
+			ok = imap_cmd_login(session, user, pass);
+		break;
+	case IMAP_AUTH_CRAM_MD5:
 		ok = imap_cmd_authenticate(session, user, pass, type);
+		break;
+	default:
+		break;
+	}
 
 	if (ok == IMAP_SUCCESS)
 		session->authenticated = TRUE;
@@ -506,25 +535,12 @@ static gint imap_auth(IMAPSession *session, const gchar *user,
 static Session *imap_session_new(PrefsAccount *account)
 {
 	IMAPSession *session;
-	SockInfo *sock;
-	gchar *pass;
 	gushort port;
 
 	g_return_val_if_fail(account != NULL, NULL);
 	g_return_val_if_fail(account->recv_server != NULL, NULL);
 	g_return_val_if_fail(account->userid != NULL, NULL);
 
-	pass = account->passwd;
-	if (!pass) {
-		gchar *tmp_pass;
-		tmp_pass = input_dialog_query_password(account->recv_server,
-						       account->userid);
-		if (!tmp_pass)
-			return NULL;
-		Xstrdup_a(pass, tmp_pass, {g_free(tmp_pass); return NULL;});
-		g_free(tmp_pass);
-	}
-
 #if USE_SSL
 	port = account->set_imapport ? account->imapport
 		: account->ssl_imap == SSL_TUNNEL ? IMAPS_PORT : IMAP4_PORT;
@@ -532,44 +548,79 @@ static Session *imap_session_new(PrefsAccount *account)
 	port = account->set_imapport ? account->imapport : IMAP4_PORT;
 #endif
 
-	log_message(_("creating IMAP4 connection to %s:%d ...\n"),
-		    account->recv_server, port);
-
-#if USE_SSL
-	if ((sock = imap_open(account->recv_server, port,
-				   account->ssl_imap)) == NULL)
-#else
-	if ((sock = imap_open(account->recv_server, port)) == NULL)
-#endif
-		return NULL;
-
 	session = g_new0(IMAPSession, 1);
 
 	session_init(SESSION(session));
 
 	SESSION(session)->type             = SESSION_IMAP;
+	SESSION(session)->sock             = NULL;
 	SESSION(session)->server           = g_strdup(account->recv_server);
-	SESSION(session)->sock             = sock;
+	SESSION(session)->port             = port;
+#if USE_SSL
+	SESSION(session)->ssl_type         = account->ssl_imap;
+#endif
 	SESSION(session)->last_access_time = time(NULL);
-	SESSION(session)->data             = NULL;
+	SESSION(session)->data             = account;
 
 	SESSION(session)->destroy          = imap_session_destroy;
 
 	session->authenticated = FALSE;
-	session->mbox      = NULL;
-	session->cmd_count = 0;
+	session->capability    = NULL;
+	session->uidplus       = FALSE;
+	session->mbox          = NULL;
+	session->cmd_count     = 0;
 
 	session_list = g_list_append(session_list, session);
 
-	if (imap_greeting(session) != IMAP_SUCCESS) {
+	if (imap_session_connect(session) != IMAP_SUCCESS) {
 		session_destroy(SESSION(session));
 		return NULL;
 	}
 
-	if (imap_cmd_capability(session) != IMAP_SUCCESS) {
-		session_destroy(SESSION(session));
-		return NULL;
+	return SESSION(session);
+}
+
+static gint imap_session_connect(IMAPSession *session)
+{
+	SockInfo *sock;
+	PrefsAccount *account;
+	gchar *pass;
+
+	g_return_val_if_fail(session != NULL, IMAP_ERROR);
+
+	account = (PrefsAccount *)(SESSION(session)->data);
+
+	log_message(_("creating IMAP4 connection to %s:%d ...\n"),
+		    SESSION(session)->server, SESSION(session)->port);
+
+	pass = account->passwd;
+	if (!pass) {
+		gchar *tmp_pass;
+		tmp_pass = input_dialog_query_password(account->recv_server,
+						       account->userid);
+		if (!tmp_pass)
+			return IMAP_ERROR;
+		Xstrdup_a(pass, tmp_pass,
+			  {g_free(tmp_pass); return IMAP_ERROR;});
+		g_free(tmp_pass);
 	}
+
+#if USE_SSL
+	if ((sock = imap_open(SESSION(session)->server, SESSION(session)->port,
+			      SESSION(session)->ssl_type)) == NULL)
+#else
+	if ((sock = imap_open(SESSION(session)->server, SESSION(session)->port))
+	    == NULL)
+#endif
+		return IMAP_ERROR;
+
+	SESSION(session)->sock = sock;
+
+	if (imap_greeting(session) != IMAP_SUCCESS)
+		return IMAP_ERROR;
+	if (imap_cmd_capability(session) != IMAP_SUCCESS)
+		return IMAP_ERROR;
+
 	if (imap_has_capability(session, "UIDPLUS"))
 		session->uidplus = TRUE;
 
@@ -581,13 +632,14 @@ static Session *imap_session_new(PrefsAccount *account)
 		ok = imap_cmd_starttls(session);
 		if (ok != IMAP_SUCCESS) {
 			log_warning(_("Can't start TLS session.\n"));
-			session_destroy(SESSION(session));
-			return NULL;
-		}
-		if (!ssl_init_socket_with_method(sock, SSL_METHOD_TLSv1)) {
-			session_destroy(SESSION(session));
-			return NULL;
+			return IMAP_ERROR;
 		}
+		if (!ssl_init_socket_with_method(sock, SSL_METHOD_TLSv1))
+			return IMAP_SOCKET;
+
+		/* capability can be changed after STARTTLS */
+		if (imap_cmd_capability(session) != IMAP_SUCCESS)
+			return IMAP_ERROR;
 	}
 #endif
 
@@ -595,11 +647,26 @@ static Session *imap_session_new(PrefsAccount *account)
 	    imap_auth(session, account->userid, pass, account->imap_auth_type)
 	    != IMAP_SUCCESS) {
 		imap_cmd_logout(session);
-		session_destroy(SESSION(session));
-		return NULL;
+		return IMAP_AUTHFAIL;
 	}
 
-	return SESSION(session);
+	return IMAP_SUCCESS;
+}
+
+static gint imap_session_reconnect(IMAPSession *session)
+{
+	g_return_val_if_fail(session != NULL, IMAP_ERROR);
+
+	session_disconnect(SESSION(session));
+
+	imap_capability_free(session);
+	session->uidplus = FALSE;
+	g_free(session->mbox);
+	session->mbox = NULL;
+	session->authenticated = FALSE;
+	SESSION(session)->state = SESSION_READY;
+
+	return imap_session_connect(session);
 }
 
 static void imap_session_destroy(Session *session)
@@ -868,6 +935,8 @@ static GSList *imap_get_msg_list(Folder *folder, FolderItem *item,
 				item->total--;
 				mlist = g_slist_remove(mlist, msginfo);
 				procmsg_msginfo_free(msginfo);
+				item->cache_dirty = TRUE;
+				item->mark_dirty = TRUE;
 				continue;
 			}
 
@@ -876,27 +945,46 @@ static GSList *imap_get_msg_list(Folder *folder, FolderItem *item,
 					item->unread++;
 					MSG_SET_PERM_FLAGS(msginfo->flags,
 							   MSG_UNREAD);
+					item->mark_dirty = TRUE;
 				}
 			} else {
-				if (MSG_IS_NEW(msginfo->flags))
+				if (MSG_IS_NEW(msginfo->flags)) {
 					item->new--;
-				if (MSG_IS_UNREAD(msginfo->flags))
+					item->mark_dirty = TRUE;
+				}
+				if (MSG_IS_UNREAD(msginfo->flags)) {
 					item->unread--;
+					item->mark_dirty = TRUE;
+				}
 				MSG_UNSET_PERM_FLAGS(msginfo->flags,
 						     MSG_NEW|MSG_UNREAD);
 			}
 
 			if (IMAP_IS_FLAGGED(imap_flags)) {
-				MSG_SET_PERM_FLAGS(msginfo->flags, MSG_MARKED);
+				if (!MSG_IS_MARKED(msginfo->flags)) {
+					MSG_SET_PERM_FLAGS(msginfo->flags,
+							   MSG_MARKED);
+					item->mark_dirty = TRUE;
+				}
 			} else {
-				MSG_UNSET_PERM_FLAGS(msginfo->flags,
-						     MSG_MARKED);
+				if (MSG_IS_MARKED(msginfo->flags)) {
+					MSG_UNSET_PERM_FLAGS(msginfo->flags,
+							     MSG_MARKED);
+					item->mark_dirty = TRUE;
+				}
 			}
 			if (IMAP_IS_ANSWERED(imap_flags)) {
-				MSG_SET_PERM_FLAGS(msginfo->flags, MSG_REPLIED);
+				if (!MSG_IS_REPLIED(msginfo->flags)) {
+					MSG_SET_PERM_FLAGS(msginfo->flags,
+							   MSG_REPLIED);
+					item->mark_dirty = TRUE;
+				}
 			} else {
-				MSG_UNSET_PERM_FLAGS(msginfo->flags,
-						     MSG_REPLIED);
+				if (MSG_IS_REPLIED(msginfo->flags)) {
+					MSG_UNSET_PERM_FLAGS(msginfo->flags,
+							     MSG_REPLIED);
+					item->mark_dirty = TRUE;
+				}
 			}
 		}
 
@@ -940,12 +1028,15 @@ static GSList *imap_get_msg_list(Folder *folder, FolderItem *item,
 			newlist = imap_get_uncached_messages(session, item,
 							     begin, last_uid,
 							     TRUE);
+			if (newlist)
+				item->cache_dirty = TRUE;
 			mlist = g_slist_concat(mlist, newlist);
 		}
 	} else {
 		imap_delete_all_cached_messages(item);
 		mlist = imap_get_uncached_messages(session, item, 0, 0, TRUE);
 		last_uid = procmsg_get_last_num_in_msg_list(mlist);
+		item->cache_dirty = TRUE;
 	}
 
 	item->mtime = uid_validity;
@@ -954,6 +1045,9 @@ static GSList *imap_get_msg_list(Folder *folder, FolderItem *item,
 
 	item->last_num = last_uid;
 
+	debug_print("cache_dirty: %d, mark_dirty: %d\n",
+		    item->cache_dirty, item->mark_dirty);
+
 catch:
 	return mlist;
 }
@@ -1673,23 +1767,29 @@ static gint imap_scan_tree_recursive(IMAPSession *session, FolderItem *item)
 
 			base = g_basename(new_item->path);
 
-			if (!folder->outbox && !strcasecmp(base, "Sent")) {
+			if (!folder->outbox &&
+			    !g_ascii_strcasecmp(base, "Sent")) {
 				new_item->stype = F_OUTBOX;
 				folder->outbox = new_item;
-			} else if (!folder->draft && !strcasecmp(base, "Drafts")) {
+			} else if (!folder->draft &&
+				   !g_ascii_strcasecmp(base, "Drafts")) {
 				new_item->stype = F_DRAFT;
 				folder->draft = new_item;
-			} else if (!folder->queue && !strcasecmp(base, "Queue")) {
+			} else if (!folder->queue &&
+				   !g_ascii_strcasecmp(base, "Queue")) {
 				new_item->stype = F_QUEUE;
 				folder->queue = new_item;
-			} else if (!folder->trash && !strcasecmp(base, "Trash")) {
+			} else if (!folder->trash &&
+				   !g_ascii_strcasecmp(base, "Trash")) {
 				new_item->stype = F_TRASH;
 				folder->trash = new_item;
 			}
 		}
 
+#if 0
 		if (new_item->no_select == FALSE)
 			imap_scan_folder(folder, new_item);
+#endif
 		if (new_item->no_sub == FALSE)
 			imap_scan_tree_recursive(session, new_item);
 	}
@@ -1948,10 +2048,9 @@ static FolderItem *imap_create_folder(Folder *folder, FolderItem *parent,
 	return new_item;
 }
 
-static gint imap_rename_folder(Folder *folder, FolderItem *item,
-			       const gchar *name)
+static gint imap_rename_folder_real(Folder *folder, FolderItem *item,
+				    FolderItem *new_parent, const gchar *name)
 {
-	gchar *dirpath;
 	gchar *newpath;
 	gchar *real_oldpath;
 	gchar *real_newpath;
@@ -1966,8 +2065,18 @@ static gint imap_rename_folder(Folder *folder, FolderItem *item,
 
 	g_return_val_if_fail(folder != NULL, -1);
 	g_return_val_if_fail(item != NULL, -1);
+	g_return_val_if_fail(folder == item->folder, -1);
 	g_return_val_if_fail(item->path != NULL, -1);
-	g_return_val_if_fail(name != NULL, -1);
+	g_return_val_if_fail(new_parent != NULL || name != NULL, -1);
+	if (new_parent) {
+		g_return_val_if_fail(item != new_parent, -1);
+		g_return_val_if_fail(item->parent != new_parent, -1);
+		g_return_val_if_fail(item->folder == new_parent->folder, -1);
+		if (g_node_is_ancestor(item->node, new_parent->node)) {
+			g_warning("folder to be moved is ancestor of new parent\n");
+			return -1;
+		}
+	}
 
 	session = imap_session_get(folder);
 	if (!session) return -1;
@@ -1984,12 +2093,30 @@ static gint imap_rename_folder(Folder *folder, FolderItem *item,
 	}
 
 	separator = imap_get_path_separator(IMAP_FOLDER(folder), item->path);
-	if (strchr(item->path, G_DIR_SEPARATOR)) {
-		dirpath = g_dirname(item->path);
-		newpath = g_strconcat(dirpath, G_DIR_SEPARATOR_S, name, NULL);
-		g_free(dirpath);
-	} else
-		newpath = g_strdup(name);
+	if (new_parent) {
+		if (name) {
+			newpath = g_strconcat(new_parent->path,
+					      G_DIR_SEPARATOR_S, name, NULL);
+		} else {
+			gchar *name_;
+
+			name_ = g_path_get_basename(item->path);
+			newpath = g_strconcat(new_parent->path,
+					      G_DIR_SEPARATOR_S, name_, NULL);
+			AUTORELEASE_STR(name_, );
+			name = name_;
+		}
+	} else {
+		if (strchr(item->path, G_DIR_SEPARATOR)) {
+			gchar *dirpath;
+
+			dirpath = g_dirname(item->path);
+			newpath = g_strconcat(dirpath, G_DIR_SEPARATOR_S, name,
+					      NULL);
+			g_free(dirpath);
+		} else
+			newpath = g_strdup(name);
+	}
 
 	real_newpath = imap_utf8_to_modified_utf7(newpath);
 	imap_path_separator_subst(real_newpath, separator);
@@ -2004,6 +2131,12 @@ static gint imap_rename_folder(Folder *folder, FolderItem *item,
 		return -1;
 	}
 
+	if (new_parent) {
+		g_node_unlink(item->node);
+		g_node_append(new_parent->node, item->node);
+		item->parent = new_parent;
+	}
+
 	g_free(item->name);
 	item->name = g_strdup(name);
 
@@ -2031,6 +2164,18 @@ static gint imap_rename_folder(Folder *folder, FolderItem *item,
 	return 0;
 }
 
+static gint imap_rename_folder(Folder *folder, FolderItem *item,
+			       const gchar *name)
+{
+	return imap_rename_folder_real(folder, item, NULL, name);
+}
+
+static gint imap_move_folder(Folder *folder, FolderItem *item,
+			     FolderItem *new_parent)
+{
+	return imap_rename_folder_real(folder, item, new_parent, NULL);
+}
+
 static gint imap_remove_folder(Folder *folder, FolderItem *item)
 {
 	gint ok;
@@ -2104,7 +2249,7 @@ static GSList *imap_get_uncached_messages(IMAPSession *session,
 	str = g_string_new(NULL);
 
 	for (;;) {
-		if ((tmp = sock_getline(SESSION(session)->sock)) == NULL) {
+		if (sock_getline(SESSION(session)->sock, &tmp) < 0) {
 			log_warning(_("error occurred while getting envelope.\n"));
 			g_string_free(str, TRUE);
 			return newlist;
@@ -2452,7 +2597,7 @@ static gchar *imap_parse_atom(IMAPSession *session, gchar *src,
 	/* read the next line if the current response buffer is empty */
 	while (isspace(*(guchar *)cur_pos)) cur_pos++;
 	while (*cur_pos == '\0') {
-		if ((nextline = sock_getline(SESSION(session)->sock)) == NULL)
+		if (sock_getline(SESSION(session)->sock, &nextline) < 0)
 			return cur_pos;
 		g_string_assign(str, nextline);
 		cur_pos = str->str;
@@ -2475,7 +2620,7 @@ static gchar *imap_parse_atom(IMAPSession *session, gchar *src,
 	} else if (*cur_pos == '{') {
 		gchar buf[32];
 		gint len;
-		gint line_len = 0;
+		gint block_len = 0;
 
 		cur_pos = strchr_cpy(cur_pos + 1, '}', buf, sizeof(buf));
 		len = atoi(buf);
@@ -2485,17 +2630,21 @@ static gchar *imap_parse_atom(IMAPSession *session, gchar *src,
 		cur_pos = str->str;
 
 		do {
-			if ((nextline = sock_getline(SESSION(session)->sock))
-				== NULL)
+			gint cur_len;
+
+			cur_len = sock_getline(SESSION(session)->sock,
+					       &nextline);
+			if (cur_len < 0)
 				return cur_pos;
-			line_len += strlen(nextline);
+			block_len += cur_len;
+			subst_null(nextline, cur_len, ' ');
 			g_string_append(str, nextline);
 			cur_pos = str->str;
 			strretchomp(nextline);
 			/* log_print("IMAP4< %s\n", nextline); */
 			debug_print("IMAP4< %s\n", nextline);
 			g_free(nextline);
-		} while (line_len < len);
+		} while (block_len < len);
 
 		memcpy(dest, cur_pos, MIN(len, dest_len - 1));
 		dest[MIN(len, dest_len - 1)] = '\0';
@@ -2529,12 +2678,16 @@ static gchar *imap_get_header(IMAPSession *session, gchar *cur_pos,
 	cur_pos = str->str;
 
 	do {
-		if ((nextline = sock_getline(SESSION(session)->sock)) == NULL)
+		gint cur_len;
+
+		cur_len = sock_getline(SESSION(session)->sock, &nextline);
+		if (cur_len < 0)
 			return cur_pos;
-		block_len += strlen(nextline);
+		block_len += cur_len;
+		subst_null(nextline, cur_len, ' ');
 		g_string_append(str, nextline);
 		cur_pos = str->str;
-		strretchomp(nextline);
+		/* strretchomp(nextline); */
 		/* debug_print("IMAP4< %s\n", nextline); */
 		g_free(nextline);
 	} while (block_len < len);
@@ -2546,7 +2699,7 @@ static gchar *imap_get_header(IMAPSession *session, gchar *cur_pos,
 
 	while (isspace(*(guchar *)cur_pos)) cur_pos++;
 	while (*cur_pos == '\0') {
-		if ((nextline = sock_getline(SESSION(session)->sock)) == NULL)
+		if (sock_getline(SESSION(session)->sock, &nextline) < 0)
 			return cur_pos;
 		g_string_assign(str, nextline);
 		cur_pos = str->str;
@@ -2570,15 +2723,16 @@ static MsgFlags imap_parse_flags(const gchar *flag_str)
 	while ((p = strchr(p, '\\')) != NULL) {
 		p++;
 
-		if (g_strncasecmp(p, "Recent", 6) == 0 && MSG_IS_UNREAD(flags)) {
+		if (g_ascii_strncasecmp(p, "Recent", 6) == 0 &&
+		    MSG_IS_UNREAD(flags)) {
 			MSG_SET_PERM_FLAGS(flags, MSG_NEW);
-		} else if (g_strncasecmp(p, "Seen", 4) == 0) {
+		} else if (g_ascii_strncasecmp(p, "Seen", 4) == 0) {
 			MSG_UNSET_PERM_FLAGS(flags, MSG_NEW|MSG_UNREAD);
-		} else if (g_strncasecmp(p, "Deleted", 7) == 0) {
+		} else if (g_ascii_strncasecmp(p, "Deleted", 7) == 0) {
 			MSG_SET_PERM_FLAGS(flags, MSG_DELETED);
-		} else if (g_strncasecmp(p, "Flagged", 7) == 0) {
+		} else if (g_ascii_strncasecmp(p, "Flagged", 7) == 0) {
 			MSG_SET_PERM_FLAGS(flags, MSG_MARKED);
-		} else if (g_strncasecmp(p, "Answered", 8) == 0) {
+		} else if (g_ascii_strncasecmp(p, "Answered", 8) == 0) {
 			MSG_SET_PERM_FLAGS(flags, MSG_REPLIED);
 		}
 	}
@@ -2594,13 +2748,13 @@ static IMAPFlags imap_parse_imap_flags(const gchar *flag_str)
 	while ((p = strchr(p, '\\')) != NULL) {
 		p++;
 
-		if (g_strncasecmp(p, "Seen", 4) == 0) {
+		if (g_ascii_strncasecmp(p, "Seen", 4) == 0) {
 			flags |= IMAP_FLAG_SEEN;
-		} else if (g_strncasecmp(p, "Deleted", 7) == 0) {
+		} else if (g_ascii_strncasecmp(p, "Deleted", 7) == 0) {
 			flags |= IMAP_FLAG_DELETED;
-		} else if (g_strncasecmp(p, "Flagged", 7) == 0) {
+		} else if (g_ascii_strncasecmp(p, "Flagged", 7) == 0) {
 			flags |= IMAP_FLAG_FLAGGED;
-		} else if (g_strncasecmp(p, "Answered", 8) == 0) {
+		} else if (g_ascii_strncasecmp(p, "Answered", 8) == 0) {
 			flags |= IMAP_FLAG_ANSWERED;
 		}
 	}
@@ -2933,7 +3087,7 @@ static gboolean imap_has_capability(IMAPSession	*session,
 	gchar **p;
 
 	for (p = session->capability; *p != NULL; ++p) {
-		if (!g_strcasecmp(*p, capability))
+		if (!g_ascii_strcasecmp(*p, capability))
 			return TRUE;
 	}
 
@@ -2942,8 +3096,10 @@ static gboolean imap_has_capability(IMAPSession	*session,
 
 static void imap_capability_free(IMAPSession *session)
 {
-	g_strfreev(session->capability);
-	session->capability = NULL;
+	if (session->capability) {
+		g_strfreev(session->capability);
+		session->capability = NULL;
+	}
 }
 
 
@@ -2967,7 +3123,8 @@ static gint imap_cmd_capability(IMAPSession *session)
 
 	capability += strlen("CAPABILITY ");
 
-	IMAP_SESSION(session)->capability = g_strsplit(capability, " ", -1);
+	imap_capability_free(session);
+	session->capability = g_strsplit(capability, " ", -1);
 
 catch:
 	ptr_array_free_strings(argbuf);
@@ -2990,7 +3147,8 @@ static gint imap_cmd_authenticate(IMAPSession *session, const gchar *user,
 	gchar *response;
 	gchar *response64;
 
-	g_return_val_if_fail(type == IMAP_AUTH_CRAM_MD5, IMAP_ERROR);
+	g_return_val_if_fail((type == 0 || type == IMAP_AUTH_CRAM_MD5),
+			     IMAP_ERROR);
 
 	auth_type = "CRAM-MD5";
 
@@ -3232,6 +3390,7 @@ static gint imap_cmd_search(IMAPSession *session, const gchar *criteria,
 	GArray *array;
 	gchar *str;
 	gchar *p, *ep;
+	gint i;
 	guint32 uid;
 
 	g_return_val_if_fail(criteria != NULL, IMAP_ERROR);
@@ -3242,20 +3401,22 @@ static gint imap_cmd_search(IMAPSession *session, const gchar *criteria,
 	imap_cmd_gen_send(session, "UID SEARCH %s", criteria);
 	if ((ok = imap_cmd_ok(session, argbuf)) != IMAP_SUCCESS) THROW(ok);
 
-	str = search_array_str(argbuf, "SEARCH");
-	if (!str) THROW(IMAP_ERROR);
-
 	array = g_array_new(FALSE, FALSE, sizeof(guint32));
 
-	p = str + strlen("SEARCH");
+	for (i = 0; i < argbuf->len; i++) {
+		str = g_ptr_array_index(argbuf, i);
+		if (strncmp(str, "SEARCH", 6) != 0)
+			continue;
 
-	while (*p != '\0') {
-		uid = strtoul(p, &ep, 10);
-		if (p < ep && uid > 0) {
-			g_array_append_val(array, uid);
-			p = ep;
-		} else
-			break;
+		p = str + 6;
+		while (*p != '\0') {
+			uid = strtoul(p, &ep, 10);
+			if (p < ep && uid > 0) {
+				g_array_append_val(array, uid);
+				p = ep;
+			} else
+				break;
+		}
 	}
 
 	*result = array;
@@ -3532,7 +3693,8 @@ static void imap_cmd_gen_send(IMAPSession *session, const gchar *format, ...)
 	session->cmd_count++;
 
 	g_snprintf(buf, sizeof(buf), "%d %s\r\n", session->cmd_count, tmp);
-	if (!strncasecmp(tmp, "LOGIN ", 6) && (p = strchr(tmp + 6, ' '))) {
+	if (!g_ascii_strncasecmp(tmp, "LOGIN ", 6) &&
+	    (p = strchr(tmp + 6, ' '))) {
 		*p = '\0';
 		log_print("IMAP4> %d %s ********\n", session->cmd_count, tmp);
 	} else
@@ -3543,7 +3705,7 @@ static void imap_cmd_gen_send(IMAPSession *session, const gchar *format, ...)
 
 static gint imap_cmd_gen_recv(IMAPSession *session, gchar **ret)
 {
-	if ((*ret = sock_getline(SESSION(session)->sock)) == NULL)
+	if (sock_getline(SESSION(session)->sock, ret) < 0)
 		return IMAP_SOCKET;
 
 	strretchomp(*ret);
@@ -3677,6 +3839,7 @@ static gchar *imap_modified_utf7_to_utf8(const gchar *mutf7_str)
 	for (p = mutf7_str; *p != '\0'; p++) {
 		/* replace: '&'  -> '+',
 			    "&-" -> '&',
+			    "+"  -> "+-",
 			    escaped ','  -> '/' */
 		if (!in_escape && *p == '&') {
 			if (*(p + 1) != '-') {
@@ -3686,6 +3849,8 @@ static gchar *imap_modified_utf7_to_utf8(const gchar *mutf7_str)
 				g_string_append_c(norm_utf7, '&');
 				p++;
 			}
+		} else if (!in_escape && *p == '+') {
+			g_string_append(norm_utf7, "+-");
 		} else if (in_escape && *p == ',') {
 			g_string_append_c(norm_utf7, '/');
 		} else if (in_escape && *p == '-') {
@@ -3696,6 +3861,8 @@ static gchar *imap_modified_utf7_to_utf8(const gchar *mutf7_str)
 		}
 	}
 
+	/* somehow iconv() returns error when the last of the string is "+-" */
+	g_string_append_c(norm_utf7, '\n');
 	norm_utf7_p = norm_utf7->str;
 	norm_utf7_len = norm_utf7->len;
 	to_len = strlen(mutf7_str) * 5;
@@ -3703,8 +3870,7 @@ static gchar *imap_modified_utf7_to_utf8(const gchar *mutf7_str)
 
 	if (iconv(cd, (ICONV_CONST gchar **)&norm_utf7_p, &norm_utf7_len,
 		  &to_p, &to_len) == -1) {
-		g_warning(_("iconv cannot convert UTF-7 to %s\n"),
-			  conv_get_locale_charset_str());
+		g_warning(_("iconv cannot convert UTF-7 to %s\n"), CS_INTERNAL);
 		g_string_free(norm_utf7, TRUE);
 		g_free(to_str);
 		return g_strdup(mutf7_str);
@@ -3714,6 +3880,7 @@ static gchar *imap_modified_utf7_to_utf8(const gchar *mutf7_str)
 	iconv(cd, NULL, NULL, &to_p, &to_len);
 	g_string_free(norm_utf7, TRUE);
 	*to_p = '\0';
+	strretchomp(to_str);
 
 	return to_str;
 }
@@ -3747,8 +3914,6 @@ static gchar *imap_utf8_to_modified_utf7(const gchar *from)
 	Xalloca(norm_utf7, norm_utf7_len + 1, return g_strdup(from));
 	norm_utf7_p = norm_utf7;
 
-#define IS_PRINT(ch) (isprint(ch) && isascii(ch))
-
 	while (from_len > 0) {
 		if (*from_tmp == '+') {
 			*norm_utf7_p++ = '+';
@@ -3756,7 +3921,7 @@ static gchar *imap_utf8_to_modified_utf7(const gchar *from)
 			norm_utf7_len -= 2;
 			from_tmp++;
 			from_len--;
-		} else if (IS_PRINT(*(guchar *)from_tmp)) {
+		} else if (g_ascii_isprint(*from_tmp)) {
 			/* printable ascii char */
 			*norm_utf7_p = *from_tmp;
 			norm_utf7_p++;
@@ -3768,7 +3933,7 @@ static gchar *imap_utf8_to_modified_utf7(const gchar *from)
 
 			/* unprintable char: convert to UTF-7 */
 			p = from_tmp;
-			while (!IS_PRINT(*(guchar *)p) && conv_len < from_len) {
+			while (!g_ascii_isprint(*p) && conv_len < from_len) {
 				conv_len += g_utf8_skip[*(guchar *)p];
 				p += g_utf8_skip[*(guchar *)p];
 			}
@@ -3778,7 +3943,7 @@ static gchar *imap_utf8_to_modified_utf7(const gchar *from)
 				  &conv_len,
 				  &norm_utf7_p, &norm_utf7_len) == -1) {
 				g_warning("iconv cannot convert %s to UTF-7\n",
-					  conv_get_locale_charset_str());
+					  CS_INTERNAL);
 				return g_strdup(from);
 			}
 
@@ -3787,8 +3952,6 @@ static gchar *imap_utf8_to_modified_utf7(const gchar *from)
 		}
 	}
 
-#undef IS_PRINT
-
 	*norm_utf7_p = '\0';
 	to_str = g_string_new(NULL);
 	for (p = norm_utf7; p < norm_utf7_p; p++) {
