@@ -1,6 +1,6 @@
 /*
  * Sylpheed -- a GTK+ based, lightweight, and fast e-mail client
- * Copyright (C) 1999-2004 Hiroyuki Yamamoto
+ * Copyright (C) 1999-2005 Hiroyuki Yamamoto
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -29,6 +29,7 @@
 #include <gtk/gtkwindow.h>
 #include <gtk/gtksignal.h>
 #include <gtk/gtkprogressbar.h>
+#include <gtk/gtkstock.h>
 #include <stdio.h>
 #include <unistd.h>
 #include <string.h>
@@ -69,17 +70,14 @@ static GList *inc_dialog_list = NULL;
 
 static guint inc_lock_count = 0;
 
-static GdkPixmap *currentxpm;
-static GdkBitmap *currentxpmmask;
-static GdkPixmap *errorxpm;
-static GdkBitmap *errorxpmmask;
-static GdkPixmap *okxpm;
-static GdkBitmap *okxpmmask;
+static GdkPixbuf *current_pixbuf;
+static GdkPixbuf *error_pixbuf;
+static GdkPixbuf *ok_pixbuf;
 
 #define MSGBUFSIZE	8192
 
 static void inc_finished		(MainWindow		*mainwin,
-					 gboolean		 new_messages);
+					 gint			 new_messages);
 static gint inc_account_mail_real	(MainWindow		*mainwin,
 					 PrefsAccount		*account);
 
@@ -145,20 +143,22 @@ static gint inc_autocheck_func			(gpointer	 data);
 /**
  * inc_finished:
  * @mainwin: Main window.
- * @new_messages: TRUE if some messages have been received.
+ * @new_messages: Number of received messages.
  * 
  * Update the folder view and the summary view after receiving
- * messages.  If @new_messages is FALSE, this function avoids unneeded
+ * messages.  If @new_messages is 0, this function avoids unneeded
  * updating.
  **/
-static void inc_finished(MainWindow *mainwin, gboolean new_messages)
+static void inc_finished(MainWindow *mainwin, gint new_messages)
 {
 	FolderItem *item;
 
+	debug_print("inc_finished(): %d new message(s)\n", new_messages);
+
 	if (prefs_common.scan_all_after_inc)
 		folderview_check_new(NULL);
 
-	if (!new_messages && !prefs_common.scan_all_after_inc) return;
+	if (new_messages <= 0 && !prefs_common.scan_all_after_inc) return;
 
 	if (prefs_common.open_inbox_on_inc) {
 		item = cur_account && cur_account->inbox
@@ -171,6 +171,23 @@ static void inc_finished(MainWindow *mainwin, gboolean new_messages)
 		if (item)
 			folderview_update_item(item, TRUE);
 	}
+
+	if (new_messages > 0 &&
+	    prefs_common.enable_newmsg_notify &&
+	    prefs_common.newmsg_notify_cmd) {
+		gchar buf[1024];
+		gchar *p;
+
+		if ((p = strchr(prefs_common.newmsg_notify_cmd, '%')) &&
+		    *(p + 1) == 'd' && !strchr(p + 2, '%'))
+			g_snprintf(buf, sizeof(buf),
+				   prefs_common.newmsg_notify_cmd,
+				   new_messages);
+		else
+			strncpy2(buf, prefs_common.newmsg_notify_cmd,
+				 sizeof(buf));
+		execute_command_line(buf, TRUE);
+	}
 }
 
 void inc_mail(MainWindow *mainwin)
@@ -206,7 +223,7 @@ void inc_mail(MainWindow *mainwin)
 		new_msgs += inc_account_mail_real(mainwin, cur_account);
 	}
 
-	inc_finished(mainwin, new_msgs > 0);
+	inc_finished(mainwin, new_msgs);
 	main_window_unlock(mainwin);
 	inc_autocheck_timer_set();
 }
@@ -255,7 +272,7 @@ gint inc_account_mail(MainWindow *mainwin, PrefsAccount *account)
 
 	new_msgs = inc_account_mail_real(mainwin, account);
 
-	inc_finished(mainwin, new_msgs > 0);
+	inc_finished(mainwin, new_msgs);
 	main_window_unlock(mainwin);
 	inc_autocheck_timer_set();
 
@@ -321,7 +338,7 @@ void inc_all_account_mail(MainWindow *mainwin, gboolean autocheck)
 		new_msgs += inc_start(inc_dialog);
 	}
 
-	inc_finished(mainwin, new_msgs > 0);
+	inc_finished(mainwin, new_msgs);
 	main_window_unlock(mainwin);
 	inc_autocheck_timer_set();
 }
@@ -344,13 +361,10 @@ static IncProgressDialog *inc_progress_dialog_create(gboolean autocheck)
 
 	progress_dialog_set_value(progress, 0.0);
 
-	gtk_widget_realize(progress->clist);
-	stock_pixmap_gdk(progress->clist, STOCK_PIXMAP_COMPLETE,
-			 &okxpm, &okxpmmask);
-	stock_pixmap_gdk(progress->clist, STOCK_PIXMAP_CONTINUE,
-			 &currentxpm, &currentxpmmask);
-	stock_pixmap_gdk(progress->clist, STOCK_PIXMAP_ERROR,
-			 &errorxpm, &errorxpmmask);
+	stock_pixbuf_gdk(progress->treeview, STOCK_PIXMAP_COMPLETE, &ok_pixbuf);
+	stock_pixbuf_gdk(progress->treeview, STOCK_PIXMAP_CONTINUE,
+			 &current_pixbuf);
+	stock_pixbuf_gdk(progress->treeview, STOCK_PIXMAP_ERROR, &error_pixbuf);
 
 	if (prefs_common.recv_dialog_mode == RECV_DIALOG_ALWAYS ||
 	    (prefs_common.recv_dialog_mode == RECV_DIALOG_MANUAL &&
@@ -377,14 +391,11 @@ static void inc_progress_dialog_set_list(IncProgressDialog *inc_dialog)
 	for (list = inc_dialog->queue_list; list != NULL; list = list->next) {
 		IncSession *session = list->data;
 		Pop3Session *pop3_session = POP3_SESSION(session->session);
-		gchar *text[3];
 
 		session->data = inc_dialog;
-
-		text[0] = NULL;
-		text[1] = pop3_session->ac_prefs->account_name;
-		text[2] = _("Standby");
-		gtk_clist_append(GTK_CLIST(inc_dialog->dialog->clist), text);
+		progress_dialog_append(inc_dialog->dialog, NULL,
+				       pop3_session->ac_prefs->account_name,
+				       _("Standby"), NULL);
 	}
 }
 
@@ -401,6 +412,8 @@ static void inc_progress_dialog_destroy(IncProgressDialog *inc_dialog)
 
 	inc_dialog_list = g_list_remove(inc_dialog_list, inc_dialog);
 
+	manage_window_destroy(inc_dialog->dialog->window, NULL);
+
 	main_window_progress_off(inc_dialog->mainwin);
 	progress_dialog_destroy(inc_dialog->dialog);
 
@@ -450,7 +463,6 @@ static void inc_session_destroy(IncSession *session)
 static gint inc_start(IncProgressDialog *inc_dialog)
 {
 	IncSession *session;
-	GtkCList *clist = GTK_CLIST(inc_dialog->dialog->clist);
 	GList *qlist;
 	Pop3Session *pop3_session;
 	IncState inc_state;
@@ -499,10 +511,12 @@ static gint inc_start(IncProgressDialog *inc_dialog)
 		qlist = next;
 	}
 
-#define SET_PIXMAP_AND_TEXT(xpm, xpmmask, str)				   \
-{									   \
-	gtk_clist_set_pixmap(clist, inc_dialog->cur_row, 0, xpm, xpmmask); \
-	gtk_clist_set_text(clist, inc_dialog->cur_row, 2, str);		   \
+#define SET_PIXMAP_AND_TEXT(pixbuf, str)				\
+{									\
+	progress_dialog_set_row_pixbuf(inc_dialog->dialog,		\
+				       inc_dialog->cur_row, pixbuf);	\
+	progress_dialog_set_row_status(inc_dialog->dialog,		\
+				       inc_dialog->cur_row, str);	\
 }
 
 	for (; inc_dialog->queue_list != NULL; inc_dialog->cur_row++) {
@@ -510,7 +524,7 @@ static gint inc_start(IncProgressDialog *inc_dialog)
 		pop3_session = POP3_SESSION(session->session);
 
 		if (pop3_session->pass == NULL) {
-			SET_PIXMAP_AND_TEXT(okxpm, okxpmmask, _("Cancelled"));
+			SET_PIXMAP_AND_TEXT(ok_pixbuf, _("Cancelled"));
 			inc_session_destroy(session);
 			inc_dialog->queue_list =
 				g_list_remove(inc_dialog->queue_list, session);
@@ -518,10 +532,10 @@ static gint inc_start(IncProgressDialog *inc_dialog)
 		}
 
 		inc_progress_dialog_clear(inc_dialog);
-		gtk_clist_moveto(clist, inc_dialog->cur_row, -1, 1.0, 0.0);
+		progress_dialog_scroll_to_row(inc_dialog->dialog,
+					      inc_dialog->cur_row);
 
-		SET_PIXMAP_AND_TEXT(currentxpm, currentxpmmask,
-				    _("Retrieving"));
+		SET_PIXMAP_AND_TEXT(current_pixbuf, _("Retrieving"));
 
 		/* begin POP3 session */
 		inc_state = inc_pop3_session_do(session);
@@ -535,33 +549,31 @@ static gint inc_start(IncProgressDialog *inc_dialog)
 					 to_human_readable(pop3_session->cur_total_recv_bytes));
 			else
 				msg = g_strdup_printf(_("Done (no new messages)"));
-			SET_PIXMAP_AND_TEXT(okxpm, okxpmmask, msg);
+			SET_PIXMAP_AND_TEXT(ok_pixbuf, msg);
 			g_free(msg);
 			break;
 		case INC_CONNECT_ERROR:
-			SET_PIXMAP_AND_TEXT(errorxpm, errorxpmmask,
+			SET_PIXMAP_AND_TEXT(error_pixbuf,
 					    _("Connection failed"));
 			break;
 		case INC_AUTH_FAILED:
-			SET_PIXMAP_AND_TEXT(errorxpm, errorxpmmask,
-					    _("Auth failed"));
+			SET_PIXMAP_AND_TEXT(error_pixbuf, _("Auth failed"));
 			break;
 		case INC_LOCKED:
-			SET_PIXMAP_AND_TEXT(errorxpm, errorxpmmask,
-					    _("Locked"));
+			SET_PIXMAP_AND_TEXT(error_pixbuf, _("Locked"));
 			break;
 		case INC_ERROR:
 		case INC_NO_SPACE:
 		case INC_IO_ERROR:
 		case INC_SOCKET_ERROR:
 		case INC_EOF:
-			SET_PIXMAP_AND_TEXT(errorxpm, errorxpmmask, _("Error"));
+			SET_PIXMAP_AND_TEXT(error_pixbuf, _("Error"));
 			break;
 		case INC_TIMEOUT:
-			SET_PIXMAP_AND_TEXT(errorxpm, errorxpmmask, _("Timeout"));
+			SET_PIXMAP_AND_TEXT(error_pixbuf, _("Timeout"));
 			break;
 		case INC_CANCEL:
-			SET_PIXMAP_AND_TEXT(okxpm, okxpmmask, _("Cancelled"));
+			SET_PIXMAP_AND_TEXT(ok_pixbuf, _("Cancelled"));
 			break;
 		default:
 			break;
@@ -851,8 +863,8 @@ static void inc_progress_dialog_set_progress(IncProgressDialog *inc_dialog,
 			   pop3_session->cur_total_num,
 			   to_human_readable
 			   (pop3_session->cur_total_recv_bytes));
-		gtk_clist_set_text(GTK_CLIST(inc_dialog->dialog->clist),
-				   inc_dialog->cur_row, 2, buf);
+		progress_dialog_set_row_status(inc_dialog->dialog,
+					       inc_dialog->cur_row, buf);
 	}
 }
 
@@ -1028,6 +1040,11 @@ static gint inc_drop_message(Pop3Session *session, const gchar *file)
 
 	if (session->ac_prefs->filter_on_recv)
 		filter_apply(prefs_common.fltlist, file, fltinfo);
+	if (!fltinfo->drop_done) {
+		if (prefs_common.enable_junk &&
+		    prefs_common.filter_junk_on_recv)
+			filter_apply(prefs_common.junk_fltlist, file, fltinfo);
+	}
 
 	if (!fltinfo->drop_done) {
 		if (folder_item_add_msg
