@@ -317,7 +317,11 @@ static void prefs_filter_create(void)
 	g_signal_connect(G_OBJECT(add_btn), "clicked",
 			 G_CALLBACK(prefs_filter_add_cb), NULL);
 
+#ifdef GTK_STOCK_EDIT
+	edit_btn = gtk_button_new_from_stock(GTK_STOCK_EDIT);
+#else
 	edit_btn = gtk_button_new_with_label(_("Edit"));
+#endif
 	gtk_widget_show(edit_btn);
 	gtk_box_pack_start(GTK_BOX(btn_hbox), edit_btn, FALSE, TRUE, 0);
 	g_signal_connect(G_OBJECT(edit_btn), "clicked",
@@ -582,9 +586,9 @@ void prefs_filter_set_msg_header_list(MsgInfo *msginfo)
 	     cur = next) {
 		next = cur->next;
 		header = (Header *)cur->data;
-		if (!g_strcasecmp(header->name, "Received") ||
-		    !g_strcasecmp(header->name, "Mime-Version") ||
-		    !g_strcasecmp(header->name, "X-UIDL")) {
+		if (!g_ascii_strcasecmp(header->name, "Received") ||
+		    !g_ascii_strcasecmp(header->name, "Mime-Version") ||
+		    !g_ascii_strcasecmp(header->name, "X-UIDL")) {
 			procheader_header_free(header);
 			rule_list_window.msg_hdr_list =
 				g_slist_remove(rule_list_window.msg_hdr_list,
