@@ -33,6 +33,7 @@
 #include "send_message.h"
 #include "procmime.h"
 #include "statusbar.h"
+#include "prefs_common.h"
 #include "prefs_filter.h"
 #include "filter.h"
 #include "folder.h"
@@ -236,8 +237,10 @@ GSList *procmsg_read_cache(FolderItem *item, gboolean scan_file)
 	}
 
 	if ((fp = procmsg_open_cache_file_with_buffer
-		(item, DATA_READ, file_buf, sizeof(file_buf))) == NULL)
+		(item, DATA_READ, file_buf, sizeof(file_buf))) == NULL) {
+		item->cache_dirty = TRUE;
 		return NULL;
+	}
 
 	debug_print("Reading summary cache...");
 
@@ -277,9 +280,10 @@ GSList *procmsg_read_cache(FolderItem *item, gboolean scan_file)
 		/* if the message file doesn't exist or is changed,
 		   don't add the data */
 		if ((type == F_MH && scan_file &&
-		     folder_item_is_msg_changed(item, msginfo)) || num == 0)
+		     folder_item_is_msg_changed(item, msginfo)) || num == 0) {
 			procmsg_msginfo_free(msginfo);
-		else {
+			item->cache_dirty = TRUE;
+		} else {
 			msginfo->folder = item;
 
 			if (!mlist)
@@ -327,6 +331,7 @@ void procmsg_set_flags(GSList *mlist, FolderItem *item)
 	if (!mark_table) {
 		item->new = item->unread = item->total = g_slist_length(mlist);
 		item->updated = TRUE;
+		item->mark_dirty = TRUE;
 		return;
 	}
 
@@ -339,6 +344,7 @@ void procmsg_set_flags(GSList *mlist, FolderItem *item)
 			if (!flags) {
 				g_hash_table_foreach(mark_table,
 						     mark_unset_new_func, NULL);
+				item->mark_dirty = TRUE;
 				break;
 			}
 		}
@@ -381,6 +387,9 @@ void procmsg_set_flags(GSList *mlist, FolderItem *item)
 	item->last_num = lastnum;
 	item->updated = TRUE;
 
+	if (unflagged > 0)
+		item->mark_dirty = TRUE;
+
 	debug_print("new: %d unread: %d unflagged: %d total: %d\n",
 		    new, unread, unflagged, total);
 
@@ -627,6 +636,7 @@ static GHashTable *procmsg_read_mark_file(FolderItem *item)
 
 	if (item->mark_queue) {
 		g_hash_table_foreach(mark_table, mark_unset_new_func, NULL);
+		item->mark_dirty = TRUE;
 	}
 
 	for (cur = item->mark_queue; cur != NULL; cur = cur->next) {
@@ -649,6 +659,7 @@ static GHashTable *procmsg_read_mark_file(FolderItem *item)
 		procmsg_write_mark_file(item, mark_table);
 		procmsg_msg_list_free(item->mark_queue);
 		item->mark_queue = NULL;
+		item->mark_dirty = FALSE;
 	}
 
 	return mark_table;
@@ -765,6 +776,24 @@ FILE *procmsg_open_mark_file(FolderItem *item, DataOpenMode mode)
 	return fp;
 }
 
+void procmsg_clear_cache(FolderItem *item)
+{
+	FILE *fp;
+
+	fp = procmsg_open_cache_file(item, DATA_WRITE);
+	if (fp)
+		fclose(fp);
+}
+
+void procmsg_clear_mark(FolderItem *item)
+{
+	FILE *fp;
+
+	fp = procmsg_open_mark_file(item, DATA_WRITE);
+	if (fp)
+		fclose(fp);
+}
+
 /* return the reversed thread tree */
 GNode *procmsg_get_thread_tree(GSList *mlist)
 {
@@ -1191,16 +1220,14 @@ void procmsg_get_filter_keyword(MsgInfo *msginfo, gchar **header, gchar **key,
 
 void procmsg_empty_trash(FolderItem *trash)
 {
-	FILE *fp;
-
 	if (trash && trash->total > 0) {
 		debug_print("Emptying messages in %s ...\n", trash->path);
 
 		folder_item_remove_all_msg(trash);
-		fp = procmsg_open_cache_file(trash, DATA_WRITE);
-		if (fp) fclose(fp);
-		fp = procmsg_open_mark_file(trash, DATA_WRITE);
-		if (fp) fclose(fp);
+		procmsg_clear_cache(trash);
+		procmsg_clear_mark(trash);
+		trash->cache_dirty = FALSE;
+		trash->mark_dirty = FALSE;
 	}
 }
 
@@ -1215,7 +1242,8 @@ void procmsg_empty_all_trash(void)
 	}
 }
 
-gint procmsg_send_queue(FolderItem *queue, gboolean save_msgs)
+gint procmsg_send_queue(FolderItem *queue, gboolean save_msgs,
+			gboolean filter_msgs)
 {
 	gint ret = 0;
 	GSList *mlist = NULL;
@@ -1231,45 +1259,68 @@ gint procmsg_send_queue(FolderItem *queue, gboolean save_msgs)
 	for (cur = mlist; cur != NULL; cur = cur->next) {
 		gchar *file;
 		MsgInfo *msginfo = (MsgInfo *)cur->data;
+		QueueInfo *qinfo;
+		gchar tmp[MAXPATHLEN + 1];
 
 		file = procmsg_get_message_file(msginfo);
-		if (file) {
-			QueueInfo *qinfo;
+		if (!file)
+			continue;
 
-			qinfo = send_get_queue_info(file);
-			if (!qinfo || send_message_queue(qinfo) < 0) {
-				g_warning(_("Sending queued message %d failed.\n"),
-					  msginfo->msgnum);
-			} else {
-				ret++;
-				if (save_msgs) {
-					FolderItem *outbox;
-					outbox = account_get_special_folder
-						(qinfo->ac, F_OUTBOX);
-					procmsg_save_to_outbox(outbox, file,
-							       TRUE);
-				}
-				folder_item_remove_msg(queue, msginfo);
-			}
+		qinfo = send_get_queue_info(file);
+		if (!qinfo || send_message_queue(qinfo) < 0) {
+			g_warning("Sending queued message %d failed.\n",
+				  msginfo->msgnum);
 			send_queue_info_free(qinfo);
 			g_free(file);
+			continue;
 		}
 
-		procmsg_msginfo_free(msginfo);
+		g_snprintf(tmp, sizeof(tmp), "%s%ctmpmsg.out.%08x",
+			   get_rc_dir(), G_DIR_SEPARATOR,
+			   (guint)random());
+
+		if (send_get_queue_contents(qinfo, tmp) == 0) {
+			if (save_msgs) {
+				FolderItem *outbox;
+				outbox = account_get_special_folder
+					(qinfo->ac, F_OUTBOX);
+				procmsg_save_to_outbox(outbox, tmp);
+			}
+			if (filter_msgs) {
+				FilterInfo *fltinfo;
+
+				fltinfo = filter_info_new();
+				fltinfo->account = qinfo->ac;
+				fltinfo->flags.perm_flags = 0;
+				fltinfo->flags.tmp_flags = MSG_RECEIVED;
+
+				filter_apply(prefs_common.fltlist, tmp,
+					     fltinfo);
+
+				filter_info_free(fltinfo);
+			}
+			unlink(tmp);
+		}
+
+		folder_item_remove_msg(queue, msginfo);
+		ret++;
+
+		send_queue_info_free(qinfo);
+		g_free(file);
 	}
 
-	g_slist_free(mlist);
+	procmsg_msg_list_free(mlist);
 
+	procmsg_clear_cache(queue);
+	queue->cache_dirty = FALSE;
 	queue->mtime = 0;
 
 	return ret;
 }
 
-gint procmsg_save_to_outbox(FolderItem *outbox, const gchar *file,
-			    gboolean is_queued)
+gint procmsg_save_to_outbox(FolderItem *outbox, const gchar *file)
 {
 	gint num;
-	FILE *fp;
 	MsgFlags flag = {0, 0};
 
 	debug_print("saving sent message...\n");
@@ -1278,43 +1329,10 @@ gint procmsg_save_to_outbox(FolderItem *outbox, const gchar *file,
 		outbox = folder_get_default_outbox();
 	g_return_val_if_fail(outbox != NULL, -1);
 
-	/* remove queueing headers */
-	if (is_queued) {
-		gchar tmp[MAXPATHLEN + 1];
-		gchar buf[BUFFSIZE];
-		FILE *outfp;
-
-		g_snprintf(tmp, sizeof(tmp), "%s%ctmpmsg.out.%08x",
-			   get_rc_dir(), G_DIR_SEPARATOR, (guint)random());
-		if ((fp = fopen(file, "rb")) == NULL) {
-			FILE_OP_ERROR(file, "fopen");
-			return -1;
-		}
-		if ((outfp = fopen(tmp, "wb")) == NULL) {
-			FILE_OP_ERROR(tmp, "fopen");
-			fclose(fp);
-			return -1;
-		}
-		while (fgets(buf, sizeof(buf), fp) != NULL)
-			if (buf[0] == '\r' || buf[0] == '\n') break;
-		while (fgets(buf, sizeof(buf), fp) != NULL)
-			fputs(buf, outfp);
-		fclose(outfp);
-		fclose(fp);
-
-		folder_item_scan(outbox);
-		if ((num = folder_item_add_msg(outbox, tmp, &flag, TRUE)) < 0) {
-			g_warning("can't save message\n");
-			unlink(tmp);
-			return -1;
-		}
-	} else {
-		folder_item_scan(outbox);
-		if ((num = folder_item_add_msg
-			(outbox, file, &flag, FALSE)) < 0) {
-			g_warning("can't save message\n");
-			return -1;
-		}
+	folder_item_scan(outbox);
+	if ((num = folder_item_add_msg(outbox, file, &flag, FALSE)) < 0) {
+		g_warning("can't save message\n");
+		return -1;
 	}
 
 	return 0;
@@ -1464,6 +1482,23 @@ MsgInfo *procmsg_msginfo_get_full_info(MsgInfo *msginfo)
 	return full_msginfo;
 }
 
+gboolean procmsg_msginfo_equal(MsgInfo *msginfo_a, MsgInfo *msginfo_b)
+{
+	if (!msginfo_a || !msginfo_b)
+		return FALSE;
+
+	if (msginfo_a == msginfo_b)
+		return TRUE;
+
+	if (msginfo_a->folder == msginfo_b->folder &&
+	    msginfo_a->msgnum == msginfo_b->msgnum &&
+	    msginfo_a->size   == msginfo_b->size   &&
+	    msginfo_a->mtime  == msginfo_b->mtime)
+		return TRUE;
+
+	return FALSE;
+}
+
 void procmsg_msginfo_free(MsgInfo *msginfo)
 {
 	if (msginfo == NULL) return;
@@ -1542,8 +1577,9 @@ static gint func_name(gconstpointer a, gconstpointer b)			\
 	if (!msginfo2->var_name)					\
 		return -1;						\
 									\
-	return strcasecmp(msginfo1->var_name, msginfo2->var_name) *	\
-		(cmp_func_sort_type == SORT_ASCENDING ? 1 : -1);	\
+	return g_ascii_strcasecmp					\
+		(msginfo1->var_name, msginfo2->var_name) *		\
+			(cmp_func_sort_type == SORT_ASCENDING ? 1 : -1);\
 }
 
 CMP_FUNC_DEF(procmsg_cmp_by_from, fromname)
