@@ -27,10 +27,14 @@
 #include <glib/gi18n.h>
 #include <gdk/gdk.h>
 #include <gdk/gdkkeysyms.h>
+#include <gdk-pixbuf/gdk-pixbuf.h>
 #include <gtk/gtkvbox.h>
 #include <gtk/gtkscrolledwindow.h>
 #include <gtk/gtksignal.h>
-#include <gdk-pixbuf/gdk-pixbuf.h>
+#include <gtk/gtkmenu.h>
+#include <gtk/gtkmenuitem.h>
+#include <gtk/gtkseparatormenuitem.h>
+#include <gtk/gtkstock.h>
 #include <stdio.h>
 #include <ctype.h>
 #include <string.h>
@@ -187,6 +191,12 @@ static gboolean textview_visibility_notify	(GtkWidget	*widget,
 static void textview_populate_popup		(GtkWidget	*widget,
 						 GtkMenu	*menu,
 						 TextView	*textview);
+static void textview_popup_menu_activate_open_uri_cb
+						(GtkMenuItem	*menuitem,
+						 gpointer	 data);
+static void textview_popup_menu_activate_add_address_cb
+						(GtkMenuItem	*menuitem,
+						 gpointer	 data);
 static void textview_popup_menu_activate_copy_cb(GtkMenuItem	*menuitem,
 						 gpointer	 data);
 static void textview_popup_menu_activate_image_cb
@@ -289,8 +299,9 @@ TextView *textview_create(void)
 	return textview;
 }
 
-static void textview_create_tags(GtkTextView *text, TextView *textview)
+static void textview_create_tags(TextView *textview)
 {
+	GtkTextView *text = GTK_TEXT_VIEW(textview->text);
 	GtkTextBuffer *buffer;
 	static PangoFontDescription *font_desc;
 
@@ -309,25 +320,32 @@ static void textview_create_tags(GtkTextView *text, TextView *textview)
 	gtk_text_buffer_create_tag(buffer, "header_title",
 				   "weight", PANGO_WEIGHT_BOLD,
 				   NULL);
-	gtk_text_buffer_create_tag(buffer, "quote0",
-				   "foreground-gdk", &quote_colors[0],
-				   NULL);
-	gtk_text_buffer_create_tag(buffer, "quote1",
-				   "foreground-gdk", &quote_colors[1],
-				   NULL);
-	gtk_text_buffer_create_tag(buffer, "quote2",
-				   "foreground-gdk", &quote_colors[2],
-				   NULL);
+
+	textview->quote0_tag =
+		gtk_text_buffer_create_tag(buffer, "quote0",
+					   "foreground-gdk", &quote_colors[0],
+					   NULL);
+	textview->quote1_tag =
+		gtk_text_buffer_create_tag(buffer, "quote1",
+					   "foreground-gdk", &quote_colors[1],
+					   NULL);
+	textview->quote2_tag =
+		gtk_text_buffer_create_tag(buffer, "quote2",
+					   "foreground-gdk", &quote_colors[2],
+					   NULL);
+	textview->link_tag =
+		gtk_text_buffer_create_tag(buffer, "link",
+					   "foreground-gdk", &uri_color,
+					   NULL);
+	textview->hover_link_tag =
+		gtk_text_buffer_create_tag(buffer, "hover-link",
+					   "foreground-gdk", &uri_color,
+					   "underline", PANGO_UNDERLINE_SINGLE,
+					   NULL);
+
 	gtk_text_buffer_create_tag(buffer, "emphasis",
 				   "foreground-gdk", &emphasis_color,
 				   NULL);
-	gtk_text_buffer_create_tag(buffer, "link",
-				   "foreground-gdk", &uri_color,
-				   NULL);
-	gtk_text_buffer_create_tag(buffer, "hover-link",
-				   "foreground-gdk", &uri_color,
-				   "underline", PANGO_UNDERLINE_SINGLE,
-				   NULL);
 #if USE_GPGME
 	gtk_text_buffer_create_tag(buffer, "good-signature",
 				   "foreground-gdk", &good_sig_color,
@@ -351,18 +369,19 @@ void textview_init(TextView *textview)
 	if (!regular_cursor)
 		regular_cursor = gdk_cursor_new(GDK_XTERM);
 
-	textview_update_message_colors();
+	textview_create_tags(textview);
+	textview_reflect_prefs(textview);
 	textview_set_all_headers(textview, FALSE);
 	textview_set_font(textview, NULL);
-	textview_create_tags(GTK_TEXT_VIEW(textview->text), textview);
 }
 
-void textview_update_message_colors(void)
+static void textview_update_message_colors(void)
 {
 	GdkColor black = {0, 0, 0, 0};
 
 	if (prefs_common.enable_color) {
-		/* grab the quote colors, converting from an int to a GdkColor */
+		/* grab the quote colors, converting from an int to a
+		   GdkColor */
 		gtkut_convert_int_to_gdk_color(prefs_common.quote_level1_col,
 					       &quote_colors[0]);
 		gtkut_convert_int_to_gdk_color(prefs_common.quote_level2_col,
@@ -377,13 +396,40 @@ void textview_update_message_colors(void)
 	}
 }
 
+static void textview_update_tags(TextView *textview)
+{
+	g_object_set(textview->quote0_tag, "foreground-gdk", &quote_colors[0],
+		     NULL);
+	g_object_set(textview->quote1_tag, "foreground-gdk", &quote_colors[1],
+		     NULL);
+	g_object_set(textview->quote2_tag, "foreground-gdk", &quote_colors[2],
+		     NULL);
+	g_object_set(textview->link_tag, "foreground-gdk", &uri_color, NULL);
+	g_object_set(textview->hover_link_tag, "foreground-gdk", &uri_color,
+		     NULL);
+}
+
+void textview_reflect_prefs(TextView *textview)
+{
+	textview_update_message_colors();
+	textview_update_tags(textview);
+	gtk_text_view_set_cursor_visible(GTK_TEXT_VIEW(textview->text),
+					 prefs_common.textview_cursor_visible);
+}
+
 void textview_show_message(TextView *textview, MimeInfo *mimeinfo,
 			   const gchar *file)
 {
+	GtkTextView *text = GTK_TEXT_VIEW(textview->text);
+	GtkTextBuffer *buffer;
+	GtkTextMark *mark;
+	GtkTextIter iter;
 	FILE *fp;
 	const gchar *charset = NULL;
 	GPtrArray *headers = NULL;
 
+	buffer = gtk_text_view_get_buffer(text);
+
 	if ((fp = fopen(file, "rb")) == NULL) {
 		FILE_OP_ERROR(file, "fopen");
 		return;
@@ -402,14 +448,9 @@ void textview_show_message(TextView *textview, MimeInfo *mimeinfo,
 	if (fseek(fp, mimeinfo->fpos, SEEK_SET) < 0) perror("fseek");
 	headers = textview_scan_header(textview, fp, charset);
 	if (headers) {
-		GtkTextView *text = GTK_TEXT_VIEW(textview->text);
-		GtkTextBuffer *buffer;
-		GtkTextIter iter;
-
 		textview_show_header(textview, headers);
 		procheader_header_array_destroy(headers);
 
-		buffer = gtk_text_view_get_buffer(text);
 		gtk_text_buffer_get_end_iter(buffer, &iter);
 		textview->body_pos = gtk_text_iter_get_offset(&iter);
 	}
@@ -419,6 +460,8 @@ void textview_show_message(TextView *textview, MimeInfo *mimeinfo,
 	fclose(fp);
 
 	textview_set_position(textview, 0);
+	mark = gtk_text_buffer_get_insert(buffer);
+	gtk_text_view_scroll_mark_onscreen(text, mark);
 }
 
 void textview_show_part(TextView *textview, MimeInfo *mimeinfo, FILE *fp)
@@ -577,13 +620,15 @@ static void textview_add_part(TextView *textview, MimeInfo *mimeinfo, FILE *fp)
 	else
 #endif
 	if (mimeinfo->filename || mimeinfo->name)
-		g_snprintf(buf, sizeof(buf), "\n[%s  %s (%d bytes)]\n",
+		g_snprintf(buf, sizeof(buf), "\n[%s  %s (%s)]\n",
 			   mimeinfo->filename ? mimeinfo->filename :
 			   mimeinfo->name,
-			   mimeinfo->content_type, mimeinfo->size);
+			   mimeinfo->content_type,
+			   to_human_readable(mimeinfo->content_size));
 	else
-		g_snprintf(buf, sizeof(buf), "\n[%s (%d bytes)]\n",
-			   mimeinfo->content_type, mimeinfo->size);
+		g_snprintf(buf, sizeof(buf), "\n[%s (%s)]\n",
+			   mimeinfo->content_type,
+			   to_human_readable(mimeinfo->content_size));
 
 #if USE_GPGME
 	if (mimeinfo->sigstatus) {
@@ -678,8 +723,8 @@ static void textview_add_parts(TextView *textview, MimeInfo *mimeinfo, FILE *fp)
 	for (;;) {
 		textview_add_part(textview, mimeinfo, fp);
 		if (mimeinfo->parent && mimeinfo->parent->content_type &&
-		    !strcasecmp(mimeinfo->parent->content_type,
-				"multipart/alternative"))
+		    !g_ascii_strcasecmp(mimeinfo->parent->content_type,
+					"multipart/alternative"))
 			mimeinfo = mimeinfo->parent->next;
 		else
 			mimeinfo = procmime_mimeinfo_next(mimeinfo);
@@ -763,7 +808,7 @@ static gboolean get_uri_part(const gchar *start, const gchar *scanpos,
 	for (ep_ = scanpos; *ep_ != '\0'; ep_++) {
 		if (!isgraph(*(const guchar *)ep_) ||
 		    !isascii(*(const guchar *)ep_) ||
-		    strchr("()<>\"", *ep_))
+		    strchr("()<>{}[]\"", *ep_))
 			break;
 	}
 
@@ -1071,7 +1116,9 @@ static void textview_write_link(TextView *textview, const gchar *str,
 	gchar *bufp;
 	RemoteURI *r_uri;
 
-	if (*str == '\0')
+	if (!str || *str == '\0')
+		return;
+	if (!uri)
 		return;
 
 	buffer = gtk_text_view_get_buffer(text);
@@ -1084,10 +1131,20 @@ static void textview_write_link(TextView *textview, const gchar *str,
 	} else
 		buf = g_strdup(str);
 
+	if (g_utf8_validate(buf, -1, NULL) == FALSE) {
+		g_free(buf);
+		return;
+	}
+
 	strcrchomp(buf);
 
-	for (bufp = buf; g_ascii_isspace(*bufp); bufp++)
-		;
+	for (bufp = buf; *bufp != '\0'; bufp = g_utf8_next_char(bufp)) {
+		gunichar ch;
+
+		ch = g_utf8_get_char(bufp);
+		if (!g_unichar_isspace(ch))
+			break;
+	}
 	if (bufp > buf)
 		gtk_text_buffer_insert(buffer, &iter, buf, bufp - buf);
 
@@ -1190,7 +1247,7 @@ static GPtrArray *textview_scan_header(TextView *textview, FILE *fp,
 		for (i = 0; i < headers->len; i++) {
 			header = g_ptr_array_index(headers, i);
 
-			if (!g_strcasecmp(header->name, dp->name)) {
+			if (!g_ascii_strcasecmp(header->name, dp->name)) {
 				if (dp->hidden)
 					procheader_header_free(header);
 				else
@@ -1239,10 +1296,10 @@ static void textview_show_header(TextView *textview, GPtrArray *headers)
 			(buffer, &iter, ":", 1,
 			 "header_title", "header", NULL);
 
-		if (!g_strcasecmp(header->name, "Subject") ||
-		    !g_strcasecmp(header->name, "From")    ||
-		    !g_strcasecmp(header->name, "To")      ||
-		    !g_strcasecmp(header->name, "Cc"))
+		if (!g_ascii_strcasecmp(header->name, "Subject") ||
+		    !g_ascii_strcasecmp(header->name, "From")    ||
+		    !g_ascii_strcasecmp(header->name, "To")      ||
+		    !g_ascii_strcasecmp(header->name, "Cc"))
 			unfold_line(header->body);
 
 		if (prefs_common.enable_color &&
@@ -1282,8 +1339,7 @@ gboolean textview_search_string(TextView *textview, const gchar *str,
 	len = g_utf8_strlen(str, -1);
 	g_return_val_if_fail(len >= 0, FALSE);
 
-	mark = gtk_text_buffer_get_insert(buffer);
-	gtk_text_buffer_get_iter_at_mark(buffer, &iter, mark);
+	gtk_text_buffer_get_selection_bounds(buffer, NULL, &iter);
 
 	if (gtkut_text_buffer_find(buffer, &iter, str, case_sens,
 				   &match_pos)) {
@@ -1292,6 +1348,7 @@ gboolean textview_search_string(TextView *textview, const gchar *str,
 		gtk_text_iter_forward_chars(&end, len);
 		/* place "insert" at the last character */
 		gtk_text_buffer_select_range(buffer, &end, &match_pos);
+		mark = gtk_text_buffer_get_insert(buffer);
 		gtk_text_view_scroll_to_mark(text, mark, 0.0, FALSE, 0.0, 0.0);
 		return TRUE;
 	}
@@ -1315,8 +1372,7 @@ gboolean textview_search_string_backward(TextView *textview, const gchar *str,
 	len = g_utf8_strlen(str, -1);
 	g_return_val_if_fail(len >= 0, FALSE);
 
-	mark = gtk_text_buffer_get_insert(buffer);
-	gtk_text_buffer_get_iter_at_mark(buffer, &iter, mark);
+	gtk_text_buffer_get_selection_bounds(buffer, &iter, NULL);
 
 	if (gtkut_text_buffer_find_backward(buffer, &iter, str, case_sens,
 					    &match_pos)) {
@@ -1324,6 +1380,7 @@ gboolean textview_search_string_backward(TextView *textview, const gchar *str,
 
 		gtk_text_iter_forward_chars(&end, len);
 		gtk_text_buffer_select_range(buffer, &match_pos, &end);
+		mark = gtk_text_buffer_get_insert(buffer);
 		gtk_text_view_scroll_to_mark(text, mark, 0.0, FALSE, 0.0, 0.0);
 		return TRUE;
 	}
@@ -1426,6 +1483,8 @@ static void textview_smooth_scroll_do(TextView *textview,
 	g_signal_emit_by_name(G_OBJECT(vadj), "value_changed", 0);
 
 	/* gdk_key_repeat_restore(); */
+
+	gtk_widget_queue_draw(GTK_WIDGET(text));
 }
 
 static void textview_smooth_scroll_one_line(TextView *textview, gboolean up)
@@ -1502,18 +1561,8 @@ static gboolean textview_smooth_scroll_page(TextView *textview, gboolean up)
 	return TRUE;
 }
 
-#warning FIXME_GTK2
-#if 0
-#define KEY_PRESS_EVENT_STOP() \
-	if (gtk_signal_n_emissions_by_name \
-		(GTK_OBJECT(widget), "key-press-event") > 0) { \
-		gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), \
-					     "key-press-event"); \
-	}
-#else
 #define KEY_PRESS_EVENT_STOP() \
 	g_signal_stop_emission_by_name(G_OBJECT(widget), "key-press-event");
-#endif
 
 static gboolean textview_key_pressed(GtkWidget *widget, GdkEventKey *event,
 				     TextView *textview)
@@ -1561,10 +1610,10 @@ static gboolean textview_key_pressed(GtkWidget *widget, GdkEventKey *event,
 		break;
 	case GDK_Escape:
 		if (summaryview && textview == messageview->textview)
-			gtk_widget_grab_focus(summaryview->ctree);
+			gtk_widget_grab_focus(summaryview->treeview);
 		else if (messageview->type == MVIEW_MIME &&
 			 textview == messageview->mimeview->textview)
-			gtk_widget_grab_focus(messageview->mimeview->ctree);
+			gtk_widget_grab_focus(messageview->mimeview->treeview);
 		break;
 	case GDK_n:
 	case GDK_N:
@@ -1689,7 +1738,7 @@ static gboolean textview_event_after(GtkWidget *widget, GdkEvent *event,
 	if (!uri)
 		return FALSE;
 
-	if (!g_strncasecmp(uri->uri, "mailto:", 7)) {
+	if (!g_ascii_strncasecmp(uri->uri, "mailto:", 7)) {
 		PrefsAccount *ac = NULL;
 		MsgInfo *msginfo = textview->messageview->msginfo;
 
@@ -1824,6 +1873,15 @@ static gboolean textview_visibility_notify(GtkWidget *widget,
 	return FALSE;
 }
 
+#define ADD_MENU(label, func)						\
+{									\
+	menuitem = gtk_menu_item_new_with_mnemonic(label);		\
+	g_signal_connect(menuitem, "activate", G_CALLBACK(func), uri);	\
+	gtk_menu_shell_append(GTK_MENU_SHELL(menu), menuitem);		\
+	g_object_set_data(G_OBJECT(menuitem), "textview", textview);	\
+	gtk_widget_show(menuitem);					\
+}
+
 static void textview_populate_popup(GtkWidget *widget, GtkMenu *menu,
 				    TextView *textview)
 {
@@ -1851,14 +1909,8 @@ static void textview_populate_popup(GtkWidget *widget, GtkMenu *menu,
 		gtk_menu_shell_append(GTK_MENU_SHELL(menu), separator);
 		gtk_widget_show(separator);
 
-		menuitem = gtk_menu_item_new_with_mnemonic
-			(_("Sa_ve this image as..."));
-		g_signal_connect
-			(menuitem, "activate",
-			 G_CALLBACK(textview_popup_menu_activate_image_cb),
-			 uri);
-		gtk_menu_shell_append(GTK_MENU_SHELL(menu), menuitem);
-		gtk_widget_show(menuitem);
+		ADD_MENU(_("Sa_ve this image as..."),
+			 textview_popup_menu_activate_image_cb);
 	}
 	on_link = textview_get_link_tag_bounds(textview, &iter, &start, &end);
 	if (!on_link)
@@ -1872,11 +1924,65 @@ static void textview_populate_popup(GtkWidget *widget, GtkMenu *menu,
 	gtk_menu_shell_append(GTK_MENU_SHELL(menu), separator);
 	gtk_widget_show(separator);
 
-	menuitem = gtk_menu_item_new_with_mnemonic(_("Copy this _link"));
-	g_signal_connect(menuitem, "activate",
-			 G_CALLBACK(textview_popup_menu_activate_copy_cb), uri);
-	gtk_menu_shell_append(GTK_MENU_SHELL(menu), menuitem);
-	gtk_widget_show(menuitem);
+	if (!g_ascii_strncasecmp(uri->uri, "mailto:", 7)) {
+		ADD_MENU(_("Compose _new message"),
+			 textview_popup_menu_activate_open_uri_cb);
+		ADD_MENU(_("Add to address _book..."),
+			 textview_popup_menu_activate_add_address_cb);
+		ADD_MENU(_("Copy this add_ress"),
+			 textview_popup_menu_activate_copy_cb);
+	} else {
+		ADD_MENU(_("_Open with Web browser"),
+			 textview_popup_menu_activate_open_uri_cb);
+		ADD_MENU(_("Copy this _link"),
+			 textview_popup_menu_activate_copy_cb);
+	}
+}
+
+static void textview_popup_menu_activate_open_uri_cb(GtkMenuItem *menuitem,
+						     gpointer data)
+{
+	RemoteURI *uri = (RemoteURI *)data;
+	TextView *textview;
+
+	g_return_if_fail(uri != NULL);
+
+	if (!uri->uri)
+		return;
+
+	textview = g_object_get_data(G_OBJECT(menuitem), "textview");
+	g_return_if_fail(textview != NULL);
+
+	if (!g_ascii_strncasecmp(uri->uri, "mailto:", 7)) {
+		PrefsAccount *ac = NULL;
+		MsgInfo *msginfo = textview->messageview->msginfo;
+
+		if (msginfo && msginfo->folder)
+			ac = account_find_from_item(msginfo->folder);
+		if (ac && ac->protocol == A_NNTP)
+			ac = NULL;
+		compose_new(ac, msginfo->folder, uri->uri + 7, NULL);
+	} else if (textview_uri_security_check(textview, uri) == TRUE)
+		open_uri(uri->uri, prefs_common.uri_cmd);
+}
+
+static void textview_popup_menu_activate_add_address_cb(GtkMenuItem *menuitem,
+							gpointer data)
+{
+	RemoteURI *uri = (RemoteURI *)data;
+	const gchar *addr;
+
+	g_return_if_fail(uri != NULL);
+
+	if (!uri->uri)
+		return;
+
+	if (!g_ascii_strncasecmp(uri->uri, "mailto:", 7))
+		addr = uri->uri + 7;
+	else
+		addr = uri->uri;
+
+	addressbook_add_contact(addr, addr, NULL);
 }
 
 static void textview_popup_menu_activate_copy_cb(GtkMenuItem *menuitem,
@@ -1891,7 +1997,7 @@ static void textview_popup_menu_activate_copy_cb(GtkMenuItem *menuitem,
 	if (!uri->uri)
 		return;
 
-	if (!g_strncasecmp(uri->uri, "mailto:", 7))
+	if (!g_ascii_strncasecmp(uri->uri, "mailto:", 7))
 		uri_string = uri->uri + 7;
 	else
 		uri_string = uri->uri;
@@ -1962,10 +2068,12 @@ static gboolean textview_uri_security_check(TextView *textview, RemoteURI *uri)
 
 		msg = g_strdup_printf(_("The real URL (%s) is different from\n"
 					"the apparent URL (%s).\n"
+					"\n"
 					"Open it anyway?"),
 				      uri->uri, visible_str);
-		aval = alertpanel(_("Warning"), msg,
-				  GTK_STOCK_YES, GTK_STOCK_NO, NULL);
+		aval = alertpanel_full(_("Fake URL warning"), msg,
+				       ALERT_WARNING, G_ALERTDEFAULT, FALSE,
+				       GTK_STOCK_YES, GTK_STOCK_NO, NULL);
 		g_free(msg);
 		if (aval == G_ALERTDEFAULT)
 			retval = TRUE;
