@@ -95,7 +95,7 @@ void hash_free_value_mem(GHashTable *table)
 
 gint str_case_equal(gconstpointer v, gconstpointer v2)
 {
-	return strcasecmp((const gchar *)v, (const gchar *)v2) == 0;
+	return g_ascii_strcasecmp((const gchar *)v, (const gchar *)v2) == 0;
 }
 
 guint str_case_hash(gconstpointer key)
@@ -142,7 +142,7 @@ gboolean str_find_equal(const gchar *haystack, const gchar *needle)
 
 gboolean str_case_find_equal(const gchar *haystack, const gchar *needle)
 {
-	return strcasecmp(haystack, needle) == 0;
+	return g_ascii_strcasecmp(haystack, needle) == 0;
 }
 
 gint to_number(const gchar *nstr)
@@ -272,7 +272,7 @@ gchar *strcasestr(const gchar *haystack, const gchar *needle)
 		return NULL;
 
 	while (haystack_len >= needle_len) {
-		if (!strncasecmp(haystack, needle, needle_len))
+		if (!g_ascii_strncasecmp(haystack, needle, needle_len))
 			return (gchar *)haystack;
 		else {
 			haystack++;
@@ -581,7 +581,7 @@ gint subject_compare_for_sort(const gchar *s1, const gchar *s2)
 	trim_subject_for_sort(str1);
 	trim_subject_for_sort(str2);
 
-	return strcasecmp(str1, str2);
+	return g_ascii_strcasecmp(str1, str2);
 }
 
 void trim_subject_for_compare(gchar *str)
@@ -592,7 +592,7 @@ void trim_subject_for_compare(gchar *str)
 	eliminate_parenthesis(str, '(', ')');
 	g_strstrip(str);
 
-	while (!strncasecmp(str, "Re:", 3)) {
+	while (!g_ascii_strncasecmp(str, "Re:", 3)) {
 		srcp = str + 3;
 		while (isspace(*srcp)) srcp++;
 		memmove(str, srcp, strlen(srcp) + 1);
@@ -605,7 +605,7 @@ void trim_subject_for_sort(gchar *str)
 
 	g_strstrip(str);
 
-	while (!strncasecmp(str, "Re:", 3)) {
+	while (!g_ascii_strncasecmp(str, "Re:", 3)) {
 		srcp = str + 3;
 		while (isspace(*srcp)) srcp++;
 		memmove(str, srcp, strlen(srcp) + 1);
@@ -619,7 +619,7 @@ void trim_subject(gchar *str)
 	gint in_brace;
 
 	destp = str;
-	while (!strncasecmp(destp, "Re:", 3)) {
+	while (!g_ascii_strncasecmp(destp, "Re:", 3)) {
 		destp += 3;
 		while (isspace(*destp)) destp++;
 	}
@@ -1048,6 +1048,17 @@ void subst_chars(gchar *str, gchar *orig, gchar subst)
 	}
 }
 
+void subst_null(gchar *str, gint len, gchar subst)
+{
+	register gchar *p = str;
+
+	while (len--) {
+		if (*p == '\0')
+			*p = subst;
+		p++;
+	}
+}
+
 void subst_for_filename(gchar *str)
 {
 	subst_chars(str, " \t\r\n\"'/\\", '_');
@@ -1122,6 +1133,32 @@ gint get_quote_level(const gchar *str)
 	return quote_level;
 }
 
+gint check_line_length(const gchar *str, gint max_chars, gint *line)
+{
+	const gchar *p = str, *q;
+	gint cur_line = 0, len;
+
+	while ((q = strchr(p, '\n')) != NULL) {
+		len = q - p + 1;
+		if (len > max_chars) {
+			if (line)
+				*line = cur_line;
+			return -1;
+		}
+		p = q + 1;
+		++cur_line;
+	}
+
+	len = strlen(p);
+	if (len > max_chars) {
+		if (line)
+			*line = cur_line;
+		return -1;
+	}
+
+	return 0;
+}
+
 gchar *strstr_with_skip_quote(const gchar *haystack, const gchar *needle)
 {
 	register guint haystack_len, needle_len;
@@ -1401,6 +1438,7 @@ GList *uri_list_extract_filenames(const gchar *uri_list)
 			while (isspace(*p)) p++;
 			if (!strncmp(p, "file:", 5)) {
 				p += 5;
+				while (*p == '/' && *(p + 1) == '/') p++;
 				q = p;
 				while (*q && *q != '\n' && *q != '\r') q++;
 
@@ -1410,6 +1448,7 @@ GList *uri_list_extract_filenames(const gchar *uri_list)
 					file = g_malloc(q - p + 2);
 					strncpy(file, p, q - p + 1);
 					file[q - p + 1] = '\0';
+					decode_uri(file, file);
 					result = g_list_append(result,file);
 				}
 			}
@@ -1451,19 +1490,19 @@ static gint axtoi(const gchar *hex_str)
 
 gboolean is_uri_string(const gchar *str)
 {
-	return (g_strncasecmp(str, "http://", 7) == 0 ||
-		g_strncasecmp(str, "https://", 8) == 0 ||
-		g_strncasecmp(str, "ftp://", 6) == 0 ||
-		g_strncasecmp(str, "www.", 4) == 0);
+	return (g_ascii_strncasecmp(str, "http://", 7) == 0 ||
+		g_ascii_strncasecmp(str, "https://", 8) == 0 ||
+		g_ascii_strncasecmp(str, "ftp://", 6) == 0 ||
+		g_ascii_strncasecmp(str, "www.", 4) == 0);
 }
 
 gchar *get_uri_path(const gchar *uri)
 {
-	if (g_strncasecmp(uri, "http://", 7) == 0)
+	if (g_ascii_strncasecmp(uri, "http://", 7) == 0)
 		return (gchar *)(uri + 7);
-	else if (g_strncasecmp(uri, "https://", 8) == 0)
+	else if (g_ascii_strncasecmp(uri, "https://", 8) == 0)
 		return (gchar *)(uri + 8);
-	else if (g_strncasecmp(uri, "ftp://", 6) == 0)
+	else if (g_ascii_strncasecmp(uri, "ftp://", 6) == 0)
 		return (gchar *)(uri + 6);
 	else
 		return (gchar *)uri;
@@ -1486,6 +1525,7 @@ gint get_uri_len(const gchar *str)
 
 /* Decodes URL-Encoded strings (i.e. strings in which spaces are replaced by
  * plusses, and escape characters are used)
+ * Note: decoded_uri and encoded_uri can point the same location
  */
 void decode_uri(gchar *decoded_uri, const gchar *encoded_uri)
 {
@@ -1514,6 +1554,17 @@ void decode_uri(gchar *decoded_uri, const gchar *encoded_uri)
 	*dec = '\0';
 }
 
+gchar *encode_uri(const gchar *filename)
+{
+	gchar *uri;
+
+	uri = g_filename_to_uri(filename, NULL, NULL);
+	if (!uri)
+		uri = g_strconcat("file://", filename, NULL);
+
+	return uri;
+}
+
 gint scan_mailto_url(const gchar *mailto, gchar **to, gchar **cc, gchar **bcc,
 		     gchar **subject, gchar **body)
 {
@@ -1554,15 +1605,16 @@ gint scan_mailto_url(const gchar *mailto, gchar **to, gchar **cc, gchar **bcc,
 
 		if (*value == '\0') continue;
 
-		if (cc && !*cc && !g_strcasecmp(field, "cc")) {
+		if (cc && !*cc && !g_ascii_strcasecmp(field, "cc")) {
 			*cc = g_strdup(value);
-		} else if (bcc && !*bcc && !g_strcasecmp(field, "bcc")) {
+		} else if (bcc && !*bcc && !g_ascii_strcasecmp(field, "bcc")) {
 			*bcc = g_strdup(value);
 		} else if (subject && !*subject &&
-			   !g_strcasecmp(field, "subject")) {
+			   !g_ascii_strcasecmp(field, "subject")) {
 			*subject = g_malloc(strlen(value) + 1);
 			decode_uri(*subject, value);
-		} else if (body && !*body && !g_strcasecmp(field, "body")) {
+		} else if (body && !*body &&
+			   !g_ascii_strcasecmp(field, "body")) {
 			*body = g_malloc(strlen(value) + 1);
 			decode_uri(*body, value);
 		}
@@ -1802,41 +1854,22 @@ gboolean file_exist(const gchar *file, gboolean allow_fifo)
 
 gboolean is_dir_exist(const gchar *dir)
 {
-	struct stat s;
-
 	if (dir == NULL)
 		return FALSE;
 
-	if (stat(dir, &s) < 0) {
-		if (ENOENT != errno) FILE_OP_ERROR(dir, "stat");
-		return FALSE;
-	}
-
-	if (S_ISDIR(s.st_mode))
-		return TRUE;
-
-	return FALSE;
+	return g_file_test(dir, G_FILE_TEST_IS_DIR);
 }
 
 gboolean is_file_entry_exist(const gchar *file)
 {
-	struct stat s;
-
 	if (file == NULL)
 		return FALSE;
 
-	if (stat(file, &s) < 0) {
-		if (ENOENT != errno) FILE_OP_ERROR(file, "stat");
-		return FALSE;
-	}
-
-	return TRUE;
+	return g_file_test(file, G_FILE_TEST_EXISTS);
 }
 
 gboolean dirent_is_regular_file(struct dirent *d)
 {
-	struct stat s;
-
 #ifdef HAVE_DIRENT_D_TYPE
 	if (d->d_type == DT_REG)
 		return TRUE;
@@ -1844,13 +1877,11 @@ gboolean dirent_is_regular_file(struct dirent *d)
 		return FALSE;
 #endif
 
-	return (stat(d->d_name, &s) == 0 && S_ISREG(s.st_mode));
+	return g_file_test(d->d_name, G_FILE_TEST_IS_REGULAR);
 }
 
 gboolean dirent_is_directory(struct dirent *d)
 {
-	struct stat s;
-
 #ifdef HAVE_DIRENT_D_TYPE
 	if (d->d_type == DT_DIR)
 		return TRUE;
@@ -1858,7 +1889,7 @@ gboolean dirent_is_directory(struct dirent *d)
 		return FALSE;
 #endif
 
-	return (stat(d->d_name, &s) == 0 && S_ISDIR(s.st_mode));
+	return g_file_test(d->d_name, G_FILE_TEST_IS_DIR);
 }
 
 gint change_dir(const gchar *dir)
@@ -2070,15 +2101,13 @@ gint remove_expired_files(const gchar *dir, guint hours)
 	return 0;
 }
 
-gint remove_dir_recursive(const gchar *dir)
+static gint remove_dir_recursive_real(const gchar *dir)
 {
 	struct stat s;
 	DIR *dp;
 	struct dirent *d;
 	gchar *prev_dir;
 
-	/* g_print("dir = %s\n", dir); */
-
 	if (stat(dir, &s) < 0) {
 		FILE_OP_ERROR(dir, "stat");
 		if (ENOENT == errno) return 0;
@@ -2097,15 +2126,6 @@ gint remove_dir_recursive(const gchar *dir)
 	prev_dir = g_get_current_dir();
 	/* g_print("prev_dir = %s\n", prev_dir); */
 
-	if (!path_cmp(prev_dir, dir)) {
-		g_free(prev_dir);
-		if (chdir("..") < 0) {
-			FILE_OP_ERROR(dir, "chdir");
-			return -1;
-		}
-		prev_dir = g_get_current_dir();
-	}
-
 	if (chdir(dir) < 0) {
 		FILE_OP_ERROR(dir, "chdir");
 		g_free(prev_dir);
@@ -2128,7 +2148,7 @@ gint remove_dir_recursive(const gchar *dir)
 		/* g_print("removing %s\n", d->d_name); */
 
 		if (dirent_is_directory(d)) {
-			if (remove_dir_recursive(d->d_name) < 0) {
+			if (remove_dir_recursive_real(d->d_name) < 0) {
 				g_warning("can't remove directory\n");
 				return -1;
 			}
@@ -2156,6 +2176,38 @@ gint remove_dir_recursive(const gchar *dir)
 	return 0;
 }
 
+gint remove_dir_recursive(const gchar *dir)
+{
+	gchar *cur_dir;
+	gint ret;
+
+	cur_dir = g_get_current_dir();
+
+	if (chdir(dir) < 0) {
+		FILE_OP_ERROR(dir, "chdir");
+		ret = -1;
+		goto leave;
+	}
+	if (chdir("..") < 0) {
+		FILE_OP_ERROR(dir, "chdir");
+		ret = -1;
+		goto leave;
+	}
+
+	ret = remove_dir_recursive_real(dir);
+
+leave:
+	if (is_dir_exist(cur_dir)) {
+		if (chdir(cur_dir) < 0) {
+			FILE_OP_ERROR(cur_dir, "chdir");
+		}
+	}
+
+	g_free(cur_dir);
+
+	return ret;
+}
+
 gint copy_file(const gchar *src, const gchar *dest, gboolean keep_backup)
 {
 	FILE *src_fp, *dest_fp;
@@ -2592,7 +2644,7 @@ gchar *get_outgoing_rfc2822_str(FILE *fp)
 	/* output header part */
 	while (fgets(buf, sizeof(buf), fp) != NULL) {
 		strretchomp(buf);
-		if (!g_strncasecmp(buf, "Bcc:", 4)) {
+		if (!g_ascii_strncasecmp(buf, "Bcc:", 4)) {
 			gint next;
 
 			for (;;) {
@@ -2606,6 +2658,11 @@ gchar *get_outgoing_rfc2822_str(FILE *fp)
 				if (fgets(buf, sizeof(buf), fp) == NULL)
 					break;
 			}
+#if 0
+		} else if (!g_ascii_strncasecmp(buf, "Date:", 5)) {
+			get_rfc822_date(buf, sizeof(buf));
+			g_string_append_printf(str, "Date: %s\r\n", buf);
+#endif
 		} else {
 			g_string_append(str, buf);
 			g_string_append(str, "\r\n");
@@ -2920,29 +2977,18 @@ gint execute_command_line(const gchar *cmdline, gboolean async)
 
 gchar *get_command_output(const gchar *cmdline)
 {
-	gchar buf[BUFFSIZE];
-	FILE *fp;
-	GString *str;
-	gchar *ret;
+	gchar *child_stdout;
+	gint status;
 
 	g_return_val_if_fail(cmdline != NULL, NULL);
 
-	if ((fp = popen(cmdline, "r")) == NULL) {
-		FILE_OP_ERROR(cmdline, "popen");
+	if (g_spawn_command_line_sync(cmdline, &child_stdout, NULL, &status,
+				      NULL) == FALSE) {
+		g_warning("Can't execute command: %s\n", cmdline);
 		return NULL;
 	}
 
-	str = g_string_new("");
-
-	while (fgets(buf, sizeof(buf), fp) != NULL)
-		g_string_append(str, buf);
-
-	pclose(fp);
-
-	ret = str->str;
-	g_string_free(str, FALSE);
-
-	return ret;
+	return child_stdout;
 }
 
 gint open_uri(const gchar *uri, const gchar *cmdline)
@@ -2993,7 +3039,7 @@ time_t remote_tzoffset_sec(const gchar *zone)
 		remoteoffset = 0;
 	} else if (strlen(zone3) == 3) {
 		for (p = ustzstr; *p != '\0'; p += 3) {
-			if (!strncasecmp(p, zone3, 3)) {
+			if (!g_ascii_strncasecmp(p, zone3, 3)) {
 				iustz = ((gint)(p - ustzstr) / 3 + 1) / 2 - 8;
 				remoteoffset = iustz * 3600;
 				break;
@@ -3115,6 +3161,13 @@ void get_rfc822_date(gchar *buf, gint len)
 		   day, dd, mon, yyyy, hh, mm, ss, tzoffset(&t));
 }
 
+/* just a wrapper to suppress the warning of gcc about %c */
+size_t my_strftime(gchar *s, size_t max, const gchar *format,
+		   const struct tm *tm)
+{
+	return strftime(s, max, format, tm);
+}
+
 static FILE *log_fp = NULL;
 
 void set_log_file(const gchar *filename)
